"""
Implements Stacked Alternating Offers (SAO) mechanism and basic negotiators.
"""
import random
import time
from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, Any, Callable, Type
from typing import Sequence, Optional, List, Tuple, Iterable, Union

import numpy as np

from negmas.common import *
from negmas.common import _ShadowAgentMechanismInterface
from negmas.events import Notification
from negmas.java import JavaCallerMixin, to_java, from_java, java_link, to_dict, JNegmasGateway, \
    python_identifier
from negmas.mechanisms import MechanismRoundResult, Mechanism
from negmas.negotiators import Negotiator, AspirationMixin, Controller
from negmas.outcomes import sample_outcomes, Outcome, outcome_is_valid, ResponseType, outcome_as_dict
from negmas.utilities import MappingUtilityFunction, normalize, UtilityFunction, UtilityValue, JavaUtilityFunction

__all__ = [
    'SAOState',
    'SAOMechanism',
    'SAOProtocol',
    'SAONegotiator',
    'RandomNegotiator',
    'LimitedOutcomesNegotiator',
    'LimitedOutcomesAcceptor',
    'AspirationNegotiator',
    'ToughNegotiator',
    'OnlyBestNegotiator',
    'SimpleTitForTatNegotiator',
    'NiceNegotiator',
    'SAOController',
    'JavaSAONegotiator'
]


@dataclass
class SAOResponse:
    """A response to an offer given by an agent in the alternating offers protocol"""
    response: ResponseType = ResponseType.NO_RESPONSE
    outcome: Optional['Outcome'] = None


@dataclass
class SAOState(MechanismState):
    current_offer: Optional['Outcome'] = None
    current_proposer: Optional[str] = None
    n_acceptances: int = 0


@dataclass
class SAOAMI(AgentMechanismInterface):
    end_on_no_response: bool = True
    publish_proposer: bool = True
    publish_n_acceptances: bool = False


class SAOMechanism(Mechanism):
    def __init__(
        self,
        issues=None,
        outcomes=None,
        n_steps=None,
        time_limit=None,
        step_time_limit=None,
        max_n_agents=None,
        dynamic_entry=True,
        keep_issue_names=True,
        cache_outcomes=True,
        max_n_outcomes: int = 1000000,
        annotation: Optional[Dict[str, Any]] = None,
        end_on_no_response=True,
        publish_proposer=True,
        publish_n_acceptances=False,
        enable_callbacks=False,
        name: Optional[str] = None,
    ):
        super().__init__(
            issues=issues,
            outcomes=outcomes,
            n_steps=n_steps,
            time_limit=time_limit,
            step_time_limit=step_time_limit,
            max_n_agents=max_n_agents,
            dynamic_entry=dynamic_entry,
            keep_issue_names=keep_issue_names,
            cache_outcomes=cache_outcomes,
            max_n_outcomes=max_n_outcomes,
            annotation=annotation,
            state_factory=SAOState,
            enable_callbacks=enable_callbacks,
            name=name,
        )
        self._current_offer = None
        self._current_proposer = None
        self._n_accepting = 0
        self._first_proposer = 0
        self._avoid_ultimatum = n_steps is not None
        self.end_negotiation_on_refusal_to_propose = end_on_no_response
        self.publish_proposer = publish_proposer
        self.publish_n_acceptances = publish_n_acceptances

    def join(self, ami: AgentMechanismInterface, state: MechanismState
             , *, ufun: Optional['UtilityFunction'] = None, role: str = 'agent') -> bool:
        if not super().join(ami, state, ufun=ufun, role=role):
            return False
        if not self.ami.dynamic_entry and not any([a.capabilities.get('propose', False) for a in self.negotiators]):
            self._current_proposer = None
            self._current_offer = None
            self._n_accepting = 0
            return False
        return True

    def extra_state(self):
        return SAOState(
            current_offer=self._current_offer
            , current_proposer=self._current_proposer.id if self._current_proposer and self.publish_proposer else None
            , n_acceptances=self._n_accepting if self.publish_n_acceptances else 0,
        )

    def round(self) -> MechanismRoundResult:
        """implements a round of the Stacked Alternating Offers Protocol.


        """
        negotiators: List[SAONegotiator] = self.negotiators
        n_negotiators = len(negotiators)

        # if this is the first step which means that there is no _current_offer
        if self.ami.state.step == 0:
            assert self._current_offer is None
            assert self._current_proposer is None

            # choose a random negotiator and set it as the current negotiator
            self._first_proposer = random.randint(0, n_negotiators - 1)
            if self._avoid_ultimatum:
                # if we are trying to avoid an ultimatum, we take an offer from everyone and ignore them but one.
                # this way, the agent cannot know its order. For example, if we have two agents and 3 steps, this will
                # be the situation after each step:
                #
                # Case 1: Assume that it ignored the offer from agent 1
                # Step, Agent 0 calls received  , Agent 1 calls received    , relative time during last call
                # 0   , counter(None)->offer1*  , counter(None) -> offer0   , 0/3
                # 1   , counter(offer2)->offer3 , counter(offer1) -> offer2 , 1/3
                # 2   , counter(offer4)->offer5 , counter(offer3) -> offer4 , 2/3
                # 3   ,                         , counter(offer5)->offer6   , 3/3
                #
                # Case 2: Assume that it ignored the offer from agent 0
                # Step, Agent 0 calls received  , Agent 1 calls received    , relative time during last call
                # 0   , counter(None)->offer1   , counter(None) -> offer0*  , 0/3
                # 1   , counter(offer0)->offer2 , counter(offer2) -> offer3 , 1/3
                # 2   , counter(offer3)->offer4 , counter(offer4) -> offer5 , 2/3
                # 3   , counter(offer5)->offer6 ,                           , 3/3
                #
                # in both cases, the agent cannot know whether its last offer going to be passed to the other agent
                # (the ultimatum scenario) or not.
                responses = []
                for neg in negotiators:
                    if not neg.capabilities.get('propose', False):
                        continue
                    strt = time.perf_counter()
                    resp = neg.counter(state=self.state, offer=None)
                    if self.ami.step_time_limit is not None and time.perf_counter() - strt > self.ami.step_time_limit:
                        return MechanismRoundResult(broken=False, timedout=True, agreement=None)
                    if self._enable_callbacks:
                        for other in self.negotiators:
                            if other is not neg:
                                other.on_partner_response(state=self.state, agent_id=neg.id
                                                          , outcome=None, response=resp)
                    if resp.response == ResponseType.END_NEGOTIATION or \
                        resp.response == ResponseType.NO_RESPONSE or (
                        resp.outcome is None and neg.capabilities.get('propose', False)):
                        return MechanismRoundResult(broken=True, timedout=False, agreement=None)
                    responses.append(resp)
                if len(responses) < 1:
                    if not self.dynamic_entry:
                        return MechanismRoundResult(broken=True, timedout=False, agreement=None, error=True
                                                , error_details='No proposers and no dynamic entry')
                    else:
                        return MechanismRoundResult(broken=False, timedout=False, agreement=None)
                resp = responses[self._first_proposer] if len(responses) > self._first_proposer else responses[0]
                neg = negotiators[self._first_proposer] if len(negotiators) > self._first_proposer else negotiators[0]
            else:
                # when there is no risk of ultimatum (n_steps is not known), we just take one first offer.
                neg = negotiators[self._first_proposer]
                strt = time.perf_counter()
                resp = neg.counter(state=self.state, offer=None)
                if self.ami.step_time_limit is not None and time.perf_counter() - strt > self.ami.step_time_limit:
                    return MechanismRoundResult(broken=False, timedout=True, agreement=None)
                if self._enable_callbacks:
                    for other in self.negotiators:
                        if other is not neg:
                            other.on_partner_response(state=self.state, agent_id=neg.id
                                                      , outcome=None, response=resp)
                if resp.response == ResponseType.END_NEGOTIATION or \
                    resp.response == ResponseType.NO_RESPONSE or resp.outcome is None:
                    return MechanismRoundResult(broken=True, timedout=False, agreement=None)
            self._n_accepting = 1
            self._current_offer = resp.outcome
            self._current_proposer = neg
            return MechanismRoundResult(broken=False, timedout=False, agreement=None)

        # this is not the first round. A round will get n_negotiators steps
        ordered_negotiators = [negotiators[(_ + self._first_proposer + 1) % n_negotiators] for _ in
                               range(n_negotiators)]
        for neg in ordered_negotiators:
            strt = time.perf_counter()
            resp = neg.counter(state=self.state, offer=self._current_offer)
            if self.ami.step_time_limit is not None and time.perf_counter() - strt > self.ami.step_time_limit:
                return MechanismRoundResult(broken=False, timedout=True, agreement=None)
            if self._enable_callbacks:
                for other in self.negotiators:
                    if other is not neg:
                        other.on_partner_response(state=self.state, agent_id=neg.id, outcome=self._current_offer
                                                  , response=resp)
            if resp.response == ResponseType.NO_RESPONSE:
                continue
            if resp.response == ResponseType.END_NEGOTIATION:
                return MechanismRoundResult(broken=True, timedout=False, agreement=None)
            if resp.response == ResponseType.ACCEPT_OFFER:
                self._n_accepting += 1
                if self._n_accepting == n_negotiators:
                    return MechanismRoundResult(broken=False, timedout=False, agreement=self._current_offer)
            if resp.response == ResponseType.REJECT_OFFER:
                proposal = resp.outcome
                if proposal is None:
                    if self.end_negotiation_on_refusal_to_propose:
                        return MechanismRoundResult(broken=True, timedout=False, agreement=None)
                    elif self._enable_callbacks:
                        for other in self.negotiators:
                            if other is not neg:
                                other.on_partner_refused_to_propose(agent_id=neg, state=self.state)
                        continue
                else:
                    self._current_offer = proposal
                    self._current_proposer = neg
                    self._n_accepting = 1
                    if self._enable_callbacks:
                        for other in self.negotiators:
                            if other is neg:
                                continue
                            other.on_partner_proposal(agent_id=neg.id, offer=proposal, state=self.state)
                    return MechanismRoundResult(broken=False, timedout=False, agreement=None)
        # we can arrive here only if all agents either refused to response or to offer.
        if not self.ami.dynamic_entry:
            raise RuntimeError('No negotiators can propose. I cannot run a meaningful negotiation')
        return MechanismRoundResult(broken=False, timedout=False, agreement=None, error=True
                                    , error_details='No negotiators can propose in a static_entry'
                                                    ' negotiation!!')


class RandomResponseMixin(object):

    def init_ranodm_response(
        self,
        p_acceptance: float = 0.15,
        p_rejection: float = 0.25,
        p_ending: float = 0.1,
    ) -> None:
        """Constructor

        Args:
            p_acceptance (float): probability of accepting offers
            p_rejection (float): probability of rejecting offers
            p_ending (float): probability of ending negotiation

        Returns:
            None

        Remarks:
            - If the summation of acceptance, rejection and ending probabilities
                is less than 1.0 then with the remaining probability a
                NO_RESPONSE is returned from respond()
        """
        if not hasattr(self, 'add_capabilities'):
            raise ValueError(
                f"self.__class__.__name__ is just a mixin for class Negotiator. You must inherit from Negotiator or"
                f" one of its descendents before inheriting from self.__class__.__name__")
        self.add_capabilities({'respond': True})
        self.p_acceptance = p_acceptance
        self.p_rejection = p_rejection
        self.p_ending = p_ending
        self.wheel: List[Tuple[float, ResponseType]] = [
            (0.0, ResponseType.NO_RESPONSE)
        ]
        if self.p_acceptance > 0.0:
            self.wheel = [
                (self.wheel[-1][0] + self.p_acceptance, ResponseType.ACCEPT_OFFER)
            ]
        if self.p_rejection > 0.0:
            self.wheel.append(
                (self.wheel[-1][0] + self.p_rejection, ResponseType.REJECT_OFFER)
            )
        if self.p_ending > 0.0:
            self.wheel.append(
                (self.wheel[-1][0] + self.p_ending, ResponseType.REJECT_OFFER)
            )
        if self.wheel[-1][0] > 1.0:
            raise ValueError('Probabilities of acceptance+rejection+ending>1')

        self.wheel = self.wheel[1:]

    # noinspection PyUnusedLocal,PyUnusedLocal
    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        r = random.random()
        for w in self.wheel:
            if w[0] >= r:
                return w[1]

        return ResponseType.NO_RESPONSE


# noinspection PyAttributeOutsideInit
class RandomProposalMixin(object):
    """The simplest possible agent.

    It just generates random offers and respond randomly to offers.
    """

    def init_random_proposal(self: Negotiator):
        self.add_capabilities(
            {
                'propose': True,
                'propose-with-value': False,
                'max-proposals': None,  # indicates infinity
            }
        )

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        if hasattr(self, '_offerable_outcomes') and self._offerable_outcomes is not None:
            return random.sample(self._offerable_outcomes, 1)[0]
        return self._ami.random_outcomes(1, astype=dict)[0]


class LimitedOutcomesAcceptorMixin(object):
    """An agent the accepts a limited set of outcomes.

    The agent accepts any of the given outcomes with the given probabilities.
    """

    def init_limited_outcomes_acceptor(self,
                                       outcomes: Optional[Union[int, Iterable['Outcome']]] = None,
                                       acceptable_outcomes: Optional[Iterable['Outcome']] = None,
                                       acceptance_probabilities: Optional[List[float]] = None,
                                       p_ending=.05,
                                       p_no_response=0.0,
                                       ) -> None:
        """Constructor

        Args:
            acceptable_outcomes (Optional[Floats]): the set of acceptable
                outcomes. If None then it is assumed to be all the outcomes of
                the negotiation.
            acceptance_probabilities (Sequence[int]): probability of accepting
                each acceptable outcome. If None then it is assumed to be unity.
            p_no_response (float): probability of refusing to respond to offers
            p_ending (float): probability of ending negotiation

        Returns:
            None

        """
        self.add_capabilities(
            {
                'respond': True,
            }
        )
        if acceptable_outcomes is not None and outcomes is None:
            raise ValueError('If you are passing acceptable outcomes explicitly then outcomes must also be passed')
        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        self.outcomes = outcomes
        if acceptable_outcomes is not None:
            acceptable_outcomes = list(acceptable_outcomes)
        self.acceptable_outcomes = acceptable_outcomes
        if acceptable_outcomes is None:
            if acceptance_probabilities is None:
                acceptance_probabilities = [0.5] * len(outcomes)
                self.acceptable_outcomes = outcomes
            else:
                self.acceptable_outcomes = outcomes
        elif acceptance_probabilities is None:
            acceptance_probabilities = [1.0] * len(acceptable_outcomes)
            if outcomes is None:
                self.outcomes = [(_,) for _ in range(len(acceptable_outcomes))]
        if self.outcomes is None:
            raise ValueError('Could not calculate all the outcomes. It is needed to assign a utility function')
        self.acceptance_probabilities = acceptance_probabilities
        u = [0.0] * len(self.outcomes)
        for p, o in zip(self.acceptance_probabilities, self.acceptable_outcomes):
            u[self.outcomes.index(o)] = p
        self.utility_function = MappingUtilityFunction(
            dict(zip(self.outcomes, u))
        )
        self.p_no_response = p_no_response
        self.p_ending = p_ending + p_no_response

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        """Respond to an offer.

        Args:
            offer (Outcome): offer being tested

        Returns:
            ResponseType: The response to the offer

        """
        ami = self._ami
        r = random.random()
        if r < self.p_no_response:
            return ResponseType.NO_RESPONSE

        if r < self.p_ending:
            return ResponseType.END_NEGOTIATION

        if self.acceptable_outcomes is None:
            if outcome_is_valid(offer, ami.issues) and random.random() < self.acceptance_probabilities:
                return ResponseType.ACCEPT_OFFER

            else:
                return ResponseType.REJECT_OFFER

        try:
            indx = self.acceptable_outcomes.index(offer)
        except ValueError:
            return ResponseType.REJECT_OFFER
        prob = self.acceptance_probabilities[indx]
        if indx < 0 or not outcome_is_valid(offer, ami.issues):
            return ResponseType.REJECT_OFFER

        if random.random() < prob:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER


class LimitedOutcomesProposerMixin(object):
    """An agent the accepts a limited set of outcomes.

    The agent proposes randomly from the given set of outcomes.

    Args:
        proposable_outcomes (Optional[Outcomes]): the set of prooposable
            outcomes. If None then it is assumed to be all the outcomes of
            the negotiation


    """

    def init_limited_outcomes_proposer(self: Negotiator, proposable_outcomes: Optional[List['Outcome']] = None) -> None:
        self.add_capabilities(
            {
                'propose': True,
                'propose-with-value': False,
                'max-proposals': None,  # indicates infinity
            }
        )
        self._offerable_outcomes = proposable_outcomes
        if proposable_outcomes is not None:
            self._offerable_outcomes = list(proposable_outcomes)

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        if self._offerable_outcomes is None:
            return self._ami.random_outcomes(1)[0]
        else:
            return random.sample(self._offerable_outcomes, 1)[0]


class LimitedOutcomesMixin(LimitedOutcomesAcceptorMixin, LimitedOutcomesProposerMixin):
    """An agent the accepts a limited set of outcomes.

    The agent accepts any of the given outcomes with the given probabilities.
    """

    def init_limited_outcomes(self,
                              outcomes: Optional[Union[int, Iterable['Outcome']]] = None,
                              acceptable_outcomes: Optional[Iterable['Outcome']] = None,
                              acceptance_probabilities: Optional[Union[float, List[float]]] = None,
                              proposable_outcomes: Optional[Iterable['Outcome']] = None,
                              p_ending=0.0,
                              p_no_response=0.0,
                              ) -> None:
        """Constructor

        Args:
            acceptable_outcomes (Optional[Outcomes]): the set of acceptable
                outcomes. If None then it is assumed to be all the outcomes of
                the negotiation.
            acceptance_probabilities (Sequence[Float]): probability of accepting
                each acceptable outcome. If None then it is assumed to be unity.
            proposable_outcomes (Optional[Outcomes]): the set of outcomes from which the agent is allowed
                to propose. If None, then it is the same as acceptable outcomes with nonzero probability
            p_no_response (float): probability of refusing to respond to offers
            p_ending (float): probability of ending negotiation

        Returns:
            None

        """
        self.init_limited_outcomes_acceptor(outcomes=outcomes, acceptable_outcomes=acceptable_outcomes
                                            , acceptance_probabilities=acceptance_probabilities
                                            , p_ending=p_ending, p_no_response=p_no_response)
        if proposable_outcomes is None and self.acceptable_outcomes is not None:
            if not isinstance(self.acceptance_probabilities, float):
                proposable_outcomes = [_ for _, p in zip(self.acceptable_outcomes, self.acceptance_probabilities)
                                       if p > 1e-9
                                       ]
        self.init_limited_outcomes_proposer(proposable_outcomes=proposable_outcomes)


class SAONegotiator(Negotiator):
    def __init__(self, assume_normalized=True, ufun: Optional[UtilityFunction] = None, name: Optional[str] = None
                 , rational_proposal=True, parent: Controller = None):
        super().__init__(name=name, ufun=ufun, parent=parent)
        self.assume_normalized = assume_normalized
        self.__end_negotiation = False
        self.my_last_proposal: Optional['Outcome'] = None
        self.my_last_proposal_utility: float = None
        self.rational_proposal = rational_proposal
        self.add_capabilities(
            {
                'respond': True,
                'propose': True,
                'max-proposals': 1,
            }
        )

    def on_notification(self, notification: Notification, notifier: str):
        if notification.type == 'end_negotiation':
            self.__end_negotiation = True

    def propose_(self, state: MechanismState) -> Optional['Outcome']:
        if not self._capabilities['propose'] or self.__end_negotiation:
            return None
        proposal = self.propose(state=state)

        # never return a proposal that is less than the reserved value
        if self.rational_proposal and self.reserved_value is not None:
            utility = None
            if proposal is not None and self.utility_function is not None:
                utility = self.utility_function(proposal)
                if utility is not None and utility < self.reserved_value:
                    return None

            if utility is not None:
                self.my_last_proposal = proposal
                self.my_last_proposal_utility = utility

        return proposal

    def respond_(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        """Respond to an offer.

        Args:
            state: `MechanismState` giving current state of the negotiation.
            offer (Outcome): offer being tested

        Returns:
            ResponseType: The response to the offer

        Remarks:
            - The default implementation never ends the negotiation except if an earler end_negotiation notification is
              sent to the negotiator
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).

        """
        if self.__end_negotiation:
            return ResponseType.END_NEGOTIATION
        return self.respond(state=state, offer=offer)

    def counter(self, state: MechanismState, offer: Optional['Outcome']) -> 'SAOResponse':
        """
        Called to counter an offer

        Args:
            state: `MechanismState` giving current state of the negotiation.
            offer: The offer to be countered. None means no offer and the agent is requested to propose an offer

        Returns:
            Tuple[ResponseType, Outcome]: The response to the given offer with a counter offer if the response is REJECT

        """
        if self.__end_negotiation:
            return SAOResponse(ResponseType.END_NEGOTIATION, None)
        if offer is None:
            return SAOResponse(ResponseType.REJECT_OFFER, self.propose_(state=state))
        response = self.respond_(state=state, offer=offer)
        if response != ResponseType.REJECT_OFFER:
            return SAOResponse(response, None)
        return SAOResponse(response, self.propose_(state=state))

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        """Respond to an offer.

        Args:
            state: `MechanismState` giving current state of the negotiation.
            offer (Outcome): offer being tested

        Returns:
            ResponseType: The response to the offer

        Remarks:
            - The default implementation never ends the negotiation
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).

        """
        if self.utility_function is None:
            return ResponseType.REJECT_OFFER
        if self.utility_function(offer) < self.reserved_value:
            return ResponseType.REJECT_OFFER
        utility = None
        if self.my_last_proposal_utility is not None:
            utility = self.my_last_proposal_utility
        if utility is None:
            myoffer = self.propose_(state=state)
            if myoffer is None:
                return ResponseType.NO_RESPONSE
            utility = self.utility_function(myoffer)
        if utility is not None and self.utility_function(offer) >= utility:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    # CALLBACK
    def on_partner_proposal(self, state: MechanismState, agent_id: str, offer: 'Outcome') -> None:
        """
        A callback called by the mechanism when a partner proposes something

        Args:
            state: `MechanismState` giving the state of the negotiation when the offer was porposed.
            agent_id: The ID of the agent who proposed
            offer: The proposal.

        Returns:
            None

        """

    def on_partner_refused_to_propose(self, state: MechanismState, agent_id: str) -> None:
        """
        A callback called by the mechanism when a partner refuses to propose

        Args:
            state: `MechanismState` giving the state of the negotiation when the partner refused to offer.
            agent_id: The ID of the agent who refused to propose

        Returns:
            None

        """

    def on_partner_response(self, state: MechanismState, agent_id: str, outcome: 'Outcome'
                            , response: 'SAOResponse') -> None:
        """
        A callback called by the mechanism when a partner responds to some offer

        Args:
            state: `MechanismState` giving the state of the negotiation when the partner responded.
            agent_id: The ID of the agent who responded
            outcome: The proposal being responded to.
            response: The response

        Returns:
            None

        """

    @abstractmethod
    def propose(self, state: MechanismState) -> Optional['Outcome']:
        """Propose a set of offers

        Args:
            state: `MechanismState` giving current state of the negotiation.

        Returns:
            The outcome being proposed or None to refuse to propose

        Remarks:
            - This function guarantees that no agents can propose something with a utility value

        """

    class Java:
        implements = ['jnegmas.sao.SAONegotiator']


class RandomNegotiator(Negotiator, RandomResponseMixin, RandomProposalMixin):
    """A negotiation agent that responds randomly in a single negotiation."""

    def __init__(
        self,
        outcomes: Union[int, List['Outcome']],
        name: str = None, parent: Controller = None,
        reserved_value: float = 0.0,
        p_acceptance=0.15,
        p_rejection=0.25,
        p_ending=0.1,
        can_propose=True
    ) -> None:
        super().__init__(name=name, parent=parent)
        # noinspection PyCallByClass
        self.init_ranodm_response(p_acceptance=p_acceptance, p_rejection=p_rejection, p_ending=p_ending)
        self.init_random_proposal()
        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        self.capabilities['propose'] = can_propose
        self.utility_function = MappingUtilityFunction(dict(zip(outcomes, np.random.rand(len(outcomes))))
                                                       , reserved_value=reserved_value)


# noinspection PyCallByClass
class LimitedOutcomesNegotiator(LimitedOutcomesMixin, SAONegotiator):
    """A negotiation agent that uses a fixed set of outcomes in a single
    negotiation."""

    def __init__(
        self, name: str = None
        , parent: Controller = None,
        outcomes: Optional[Union[int, Iterable['Outcome']]] = None,
        acceptable_outcomes: Optional[List['Outcome']] = None,
        acceptance_probabilities: Optional[Union[float, List[float]]] = None,
        p_ending=0.0,
        p_no_response=0.0,
    ) -> None:
        super().__init__(name=name, parent=parent)
        self.init_limited_outcomes(p_ending=p_ending, p_no_response=p_no_response
                                   , acceptable_outcomes=acceptable_outcomes
                                   , acceptance_probabilities=acceptance_probabilities, outcomes=outcomes)


# noinspection PyCallByClass
class LimitedOutcomesAcceptor(SAONegotiator, LimitedOutcomesAcceptorMixin):
    """A negotiation agent that uses a fixed set of outcomes in a single
    negotiation."""

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        return LimitedOutcomesAcceptorMixin.respond(self, state=state, offer=offer)

    def __init__(
        self,
        name: str = None, parent: Controller = None,
        outcomes: Optional[Union[int, Iterable['Outcome']]] = None,
        acceptable_outcomes: Optional[List['Outcome']] = None,
        acceptance_probabilities: Optional[List[float]] = None,
        p_ending=0.0,
        p_no_response=0.0,
    ) -> None:
        SAONegotiator.__init__(self, name=name, parent=parent)
        self.init_limited_outcomes_acceptor(
            p_ending=p_ending,
            p_no_response=p_no_response,
            acceptable_outcomes=acceptable_outcomes,
            acceptance_probabilities=acceptance_probabilities,
            outcomes=outcomes
        )
        self.add_capabilities(
            {
                'propose': False,
            }
        )

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        return None


class AspirationNegotiator(SAONegotiator, AspirationMixin):
    def __init__(self, name=None, ufun=None, parent: Controller = None, max_aspiration=0.95, aspiration_type='boulware'
                 , dynamic_ufun=True, randomize_offer=False, can_propose=True, assume_normalized=False):
        super().__init__(name=name, assume_normalized=assume_normalized, parent=parent, ufun=ufun)
        self.aspiration_init(max_aspiration=max_aspiration, aspiration_type=aspiration_type)
        self.ordered_outcomes = []
        self.dynamic_ufun = dynamic_ufun
        self.randomize_offer = randomize_offer
        self._max_aspiration = self.max_aspiration
        self.ufun_max = None
        self.ufun_min = None
        self.__ufun_modified = False
        self.add_capabilities(
            {
                'respond': True,
                'propose': can_propose,
                'propose-with-value': False,
                'max-proposals': None,  # indicates infinity
            }
        )

    @property
    def eu(self) -> Callable[['Outcome'], Optional['UtilityValue']]:
        """
        The utility function in the given negotiation taking opponent model into account.

        Remarks:
            - If no utility_function is internally stored, `eu` still returns a valid callable that returns None for
              everything.
        """
        if hasattr(self, 'opponent_model'):
            return lambda x: self.utility_function(x) * self.opponent_model.probability_of_acceptance(x)
        else:
            return self.utility_function

    def _update_ordered_outcomes(self):
        outcomes = self._ami.discrete_outcomes()
        self.ordered_outcomes = sorted([(self.utility_function(outcome), outcome) for outcome in outcomes]
                                       , key=lambda x: x[0], reverse=True)
        if not self.assume_normalized:
            self.ufun_max = self.ordered_outcomes[0][0]
            self.ufun_min = self.ordered_outcomes[-1][0]
            if self.reserved_value is not None and self.ufun_min < self.reserved_value:
                self.ufun_min = self.reserved_value

    def on_notification(self, notification: Notification, notifier: str):
        super().on_notification(notification, notifier)
        if notification.type == 'ufun_modified':
            if self.dynamic_ufun:
                self.__ufun_modified = True
                self._update_ordered_outcomes()

    def join(self, ami: AgentMechanismInterface, state: MechanismState
             , *, ufun: Optional['UtilityFunction'] = None, role: str = 'agent') -> bool:
        if not super().join(ami, state, ufun=ufun, role=role):
            return False
        self.__ufun_modified = False
        self._update_ordered_outcomes()
        return True

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        if self.utility_function is None:
            return ResponseType.REJECT_OFFER
        u = self.utility_function(offer)
        if u is None:
            return ResponseType.REJECT_OFFER
        asp = self.aspiration(state.relative_time) * (self.ufun_max - self.ufun_min) + self.ufun_min
        if u >= asp and u >= self.reserved_value:
            return ResponseType.ACCEPT_OFFER
        if asp < self.reserved_value:
            return ResponseType.END_NEGOTIATION
        return ResponseType.REJECT_OFFER

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        asp = self.aspiration(state.relative_time) * (self.ufun_max - self.ufun_min) + self.ufun_min
        if asp < self.reserved_value:
            return None
        for i, (u, o) in enumerate(self.ordered_outcomes):
            if u is None:
                continue
            if u < asp:
                if u < self.reserved_value:
                    return None
                if i == 0:
                    return self.ordered_outcomes[i][1]
                if self.randomize_offer:
                    return random.sample(self.ordered_outcomes[:i], 1)[0][1]
                return self.ordered_outcomes[i - 1][1]
        if self.randomize_offer:
            return random.sample(self.ordered_outcomes, 1)[0][1]
        return self.ordered_outcomes[-1][1]


class NiceNegotiator(SAONegotiator, RandomProposalMixin):

    def __init__(self, *args, **kwargs):
        SAONegotiator.__init__(self, *args, **kwargs)
        self.init_random_proposal()

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        return ResponseType.ACCEPT_OFFER

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        return RandomProposalMixin.propose(self=self, state=state)


class ToughNegotiator(SAONegotiator):
    def __init__(self, name=None, parent: Controller = None, dynamic_ufun=True, can_propose=True):
        super().__init__(name=name, parent=parent)
        self.best_outcome = None
        self.dynamic_ufun = dynamic_ufun
        self._offerable_outcomes = None
        self.add_capabilities(
            {
                'respond': True,
                'propose': can_propose,
                'propose-with-value': False,
                'max-proposals': None,  # indicates infinity
            }
        )

    @property
    def eu(self) -> Callable[['Outcome'], Optional['UtilityValue']]:
        """
        The utility function in the given negotiation taking opponent model into account.

        Remarks:
            - If no utility_function is internally stored, `eu` still returns a valid callable that returns None for
              everything.
        """
        if hasattr(self, 'opponent_model'):
            return lambda x: self.utility_function(x) * self.opponent_model.probability_of_acceptance(x)
        else:
            return self.utility_function

    def join(self, ami: AgentMechanismInterface, state: MechanismState
             , *, ufun: Optional['UtilityFunction'] = None, role: str = 'agent') -> bool:
        if not super().join(ami, state, ufun=ufun, role=role):
            return False
        outcomes = self._ami.outcomes if self._offerable_outcomes is None else self._offerable_outcomes
        if self.utility_function is None:
            return True
        self.best_outcome = max([(self.utility_function(outcome), outcome) for outcome in outcomes])[1]
        return True

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        if self.dynamic_ufun:
            outcomes = self._ami.outcomes if self._offerable_outcomes is None else self._offerable_outcomes
            if self.utility_function is None:
                return ResponseType.REJECT_OFFER
            self.best_outcome = max([(self.utility_function(outcome), outcome) for outcome in outcomes])[1]
        if offer == self.best_outcome:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        if not self._capabilities['propose']:
            return None
        if self.dynamic_ufun:
            outcomes = self._ami.outcomes if self._offerable_outcomes is None else self._offerable_outcomes
            self.best_outcome = max([(self.eu(outcome), outcome) for outcome in outcomes])[1]
        return self.best_outcome


class OnlyBestNegotiator(SAONegotiator):
    def __init__(self, name=None, parent: Controller = None, dynamic_ufun=True
                 , min_utility=0.95, top_fraction=0.05, best_first=False
                 , probabilisic_offering=True, can_propose=True
                 ):
        super().__init__(name=name, parent=parent)
        self._offerable_outcomes = None
        self.best_outcome = []
        self.dynamic_ufun = dynamic_ufun
        self.top_fraction = top_fraction
        self.min_utility = min_utility
        self.best_first = best_first
        self.probabilisit_offering = probabilisic_offering
        self.ordered_outcomes = []
        self.acceptable_outcomes = []
        self.wheel = np.array([])
        self.offered = set([])
        self.add_capabilities(
            {
                'respond': True,
                'propose': can_propose,
                'propose-with-value': False,
                'max-proposals': None,  # indicates infinity
            }
        )

    @property
    def eu(self) -> Callable[['Outcome'], Optional['UtilityValue']]:
        """
        The utility function in the given negotiation taking opponent model into account.

        Remarks:
            - If no utility_function is internally stored, `eu` still returns a valid callable that returns None for
              everything.
        """
        if hasattr(self, 'opponent_model'):
            return lambda x: self.utility_function(x) * self.opponent_model.probability_of_acceptance(x)
        else:
            return self.utility_function

    def acceptable(self):
        outcomes = self._ami.outcomes if self._offerable_outcomes is None else self._offerable_outcomes
        eu_outcome = [(self.eu(outcome), outcome) for outcome in outcomes]
        self.ordered_outcomes = sorted(eu_outcome
                                       , key=lambda x: x[0], reverse=True)
        if self.min_utility is None:
            selected, selected_utils = [], []
        else:
            util_limit = self.min_utility * self.ordered_outcomes[0][0]
            selected, selected_utils = [], []
            for u, o in self.ordered_outcomes:
                if u >= util_limit:
                    selected.append(o)
                    selected_utils.append(u)
                else:
                    break
        if self.top_fraction is not None:
            frac_limit = max(1, int(round(self.top_fraction * len(self.ordered_outcomes))))
        else:
            frac_limit = len(outcomes)

        if frac_limit >= len(selected) > 0:
            sum = np.asarray(selected_utils).sum()
            if sum > 0.0:
                selected_utils /= sum
                selected_utils = np.cumsum(selected_utils)
            else:
                selected_utils = np.linspace(0.0, 1.0, len(selected_utils))
            return selected, selected_utils
        if frac_limit > 0:
            n_sel = len(selected)
            fsel = [_[1] for _ in self.ordered_outcomes[n_sel:frac_limit]]
            futil = [_[0] for _ in self.ordered_outcomes[n_sel:frac_limit]]
            selected_utils = selected_utils + futil
            sum = np.asarray(selected_utils).sum()
            if sum > 0.0:
                selected_utils /= sum
                selected_utils = np.cumsum(selected_utils)
            else:
                selected_utils = np.linspace(0.0, 1.0, len(selected_utils))
            return selected + fsel, selected_utils
        return [], []

    def join(self, ami: AgentMechanismInterface, state: MechanismState
             , *, ufun: Optional['UtilityFunction'] = None, role: str = 'agent') -> bool:
        if not super().join(ami, state, ufun=ufun, role=role):
            return False
        self.acceptable_outcomes, self.wheel = self.acceptable()
        return True

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        if self.dynamic_ufun:
            self.acceptable_outcomes, self.wheel = self.acceptable()
        if offer in self.acceptable_outcomes:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        if not self._capabilities['propose']:
            return None
        if self.dynamic_ufun:
            self.acceptable_outcomes, self.wheel = self.acceptable()
        if self.best_first:
            for o in self.acceptable_outcomes:
                if o not in self.offered:
                    self.offered.add(o)
                    return o
        if len(self.acceptable_outcomes) > 0:
            if self.probabilisit_offering:
                r = random.random()
                for o, w in zip(self.acceptable_outcomes, self.wheel):
                    if w > r:
                        return o
                return random.sample(self.acceptable_outcomes, 1)[0]
            return random.sample(self.acceptable_outcomes, 1)[0]
        return None


class SimpleTitForTatNegotiator(SAONegotiator):
    """Implements a generalized tit-for-tat strategy"""

    def __init__(self, name: str = None, parent: Controller = None
                 , ufun: Optional['UtilityFunction'] = None, kindness=0.0
                 , randomize_offer=False, initial_concession: Union[float, str] = 'min'):
        super().__init__(name=name, ufun=ufun, parent=parent)
        self._offerable_outcomes = None
        self.received_utilities = []
        self.proposed_utility = None
        self.kindness = kindness
        self.ordered_outcomes = None
        self.initial_concession = initial_concession
        self.randomize_offer = randomize_offer

    def join(self, ami: AgentMechanismInterface, state: MechanismState
             , *, ufun: Optional['UtilityFunction'] = None, role: str = 'agent') -> bool:
        if not super().join(ami, state, ufun=ufun, role=role):
            return False
        outcomes = self._ami.outcomes if self._offerable_outcomes is None else self._offerable_outcomes
        if outcomes is None:
            outcomes = sample_outcomes(self._ami.issues, keep_issue_names=True)
        if self.utility_function is None:
            return True
        self.ordered_outcomes = sorted([(self.utility_function(outcome), outcome) for outcome in outcomes]
                                       , key=lambda x: x[0], reverse=True)
        return True

    def respond(self, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        my_offer = self.propose(state=state)
        if self.utility_function is None:
            return ResponseType.REJECT_OFFER
        u = self.utility_function(offer)
        if len(self.received_utilities) < 2:
            self.received_utilities.append(u)
        else:
            self.received_utilities[-1] = u
        if u >= self.utility_function(my_offer):
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    def _outcome_at_utility(self, asp: float, n: int) -> List['Outcome']:
        for i, (u, o) in enumerate(self.ordered_outcomes):
            if u is None:
                continue
            if u < asp:
                if i == 0:
                    return [self.ordered_outcomes[0][1]] * n
                if self.randomize_offer:
                    if n < i:
                        return [random.sample(self.ordered_outcomes, 1)[0][1]] * n
                    else:
                        return [_[1] for _ in self.ordered_outcomes[: i]]
                else:
                    return [self.ordered_outcomes[i][1]] * n

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        if len(self.received_utilities) < 2 or self.proposed_utility is None:
            if len(self.received_utilities) < 1:
                return self.ordered_outcomes[0][1]
            if isinstance(self.initial_concession, str) and self.initial_concession == 'min':
                asp = None
                for u, o in self.ordered_outcomes:
                    if u is None:
                        continue
                    if asp is not None and u < asp:
                        break
                    asp = u
                if asp is None:
                    return self.ordered_outcomes[0][1]
            else:
                asp = self.ordered_outcomes[0][0] * (1 - self.initial_concession)
            return self._outcome_at_utility(asp=asp, n=1)[0]
        opponent_concession = self.received_utilities[1] - self.received_utilities[0]
        asp = self.proposed_utility - opponent_concession * (1 + self.kindness)
        return self._outcome_at_utility(asp=asp, n=1)[0]


def _to_java_response(response: ResponseType) -> int:
    if response == ResponseType.ACCEPT_OFFER:
        return 0
    if response == ResponseType.REJECT_OFFER:
        return 1
    if response == ResponseType.END_NEGOTIATION:
        return 2
    if response == ResponseType.NO_RESPONSE:
        return 3
    raise ValueError(f'Unknown response{response}')


def _from_java_response(response: int) -> ResponseType:
    if response == 0:
        return ResponseType.ACCEPT_OFFER
    if response == 1:
        return ResponseType.REJECT_OFFER
    if response == 2:
        return ResponseType.END_NEGOTIATION
    if response == 3:
        return ResponseType.NO_RESPONSE
    raise ValueError(f'Unknown response type {response} returned from the Java underlying negotiator')


class JavaSAONegotiator(SAONegotiator, JavaCallerMixin):

    def __init__(self, java_object, java_class_name: Optional[str]
                 , auto_load_java: bool = False, outcome_type: Type = dict):
        if java_class_name is not None or java_object is not None:
            self.init_java_bridge(java_object=java_object, java_class_name=java_class_name
                                  , auto_load_java=auto_load_java, python_shadow_object=None)
            if java_object is None:
                self._java_object.fromMap(to_java(self))
        d = {python_identifier(k): v for k, v in JNegmasGateway.gateway.entry_point.toMap(self._java_object).items()}
        ufun = d.get('utility_function', None)
        ufun = JavaUtilityFunction(ufun, None) if ufun is not None else None
        super().__init__(name=d.get('name', None), assume_normalized=d.get('assume_normalized', False)
                         , ufun=ufun
                         , rational_proposal=d.get('rational_proposal', True), parent=d.get('parent', None))
        self._outcome_type = outcome_type
        self.add_capabilities(
            {
                'respond': True,
                'propose': True,
                'propose-with-value': False,
                'max-proposals': None,  # indicates infinity
            }
        )

    def on_partner_proposal(self, state: MechanismState, agent_id: str, offer: 'Outcome') -> None:
        self._java_object.onPartnerProposal(to_java(state), agent_id, to_java(offer))

    def on_partner_refused_to_propose(self, state: MechanismState, agent_id: str) -> None:
        self._java_object.onPartnerRefusedToPropose(to_java(state), agent_id)

    def on_partner_response(self, state: MechanismState, agent_id: str, outcome: 'Outcome',
                            response: 'SAOResponse') -> None:
        self._java_object.onPartnerResponse(to_java(state), agent_id, to_java(outcome),
                                            self._to_java_response(response))

    def isin(self, negotiation_id: Optional[str]) -> bool:
        return self._java_object.isIn(negotiation_id)

    def join(self, ami: AgentMechanismInterface, state: MechanismState, *,
             ufun: Optional['UtilityFunction'] = None, role: str = 'agent') -> bool:
        return self._java_object.join(java_link(_ShadowAgentMechanismInterface(ami)), to_java(state), ufun, role)

    def on_negotiation_start(self, state: MechanismState) -> None:
        self._java_object.onNegotiationStart(to_java(state))

    def on_round_start(self, state: MechanismState) -> None:
        super().on_round_start(state)

    def on_mechanism_error(self, state: MechanismState) -> None:
        super().on_mechanism_error(state)

    def on_round_end(self, state: MechanismState) -> None:
        super().on_round_end(state)

    def on_leave(self, state: MechanismState) -> None:
        super().on_leave(state)

    def on_negotiation_end(self, state: MechanismState) -> None:
        super().on_negotiation_end(state)

    @classmethod
    def from_dict(cls, java_object, *args, parent: Controller = None) -> 'JavaSAONegotiator':
        """Creates a Java negotiator from an object returned from the JVM implementing PySAONegotiator"""
        ufun = java_object.getUtilityFunction()
        if ufun is not None:
            ufun = JavaUtilityFunction.from_dict(java_object=ufun)
        return JavaCallerMixin.from_dict(java_object, name=java_object.getName()
                                         , assume_normalized=java_object.getAssumeNormalized()
                                         , rational_proposal=java_object.getRationalProposal()
                                         , parent=parent
                                         , ufun=ufun)

    def on_notification(self, notification: Notification, notifier: str):
        super().on_notification(notification=notification, notifier=notifier)
        jnotification = {'type': notification.type, 'data': to_java(notification.data)}
        self._java_object.on_notification(jnotification, notifier)

    def respond(self, state: MechanismState, offer: 'Outcome'):
        return _from_java_response(self._java_object.respond(to_java(state), outcome_as_dict(offer)))

    def propose(self, state: MechanismState) -> Optional['Outcome']:
        outcome = from_java(self._java_object.propose(to_java(state)))
        if outcome is None:
            return None
        if self._outcome_type == dict:
            return outcome
        if self._outcome_type == tuple:
            return tuple(outcome.values())
        return self._outcome_type(outcome)

    # class Java:
    #    implements = ['jnegmas.sao.SAONegotiator']


class _ShadowSAONegotiator:
    """A python shadow to a java negotiator"""

    class Java:
        implements = ['jnegmas.sao.SAONegotiator']

    def to_java(self):
        return to_dict(self.shadow)

    def __init__(self, negotiator: SAONegotiator):
        self.shadow = negotiator

    def respond(self, state, outcome):
        return _to_java_response(self.shadow.respond(from_java(state), from_java(outcome)))

    def propose(self, state):
        return to_java(self.shadow.propose(from_java(state)))

    def isIn(self, negotiation_id):
        return to_java(self.shadow.isin(negotiation_id=negotiation_id))

    def join(self, ami, state, ufun, role):
        return to_java(self.shadow.join(ami=from_java(ami), state=from_java(state)
                                        , ufun=JavaUtilityFunction(ufun, None) if ufun is not None else None
                                        , role=role))

    def onNegotiationStart(self, state):
        return to_java(self.shadow.on_negotiation_start(from_java(state)))

    def onNegotiationEnd(self, state):
        return to_java(self.shadow.on_negotiation_end(from_java(state)))

    def onMechanismError(self, state):
        return to_java(self.shadow.on_mechanism_error(from_java(state)))

    def onRoundStart(self, state):
        return to_java(self.shadow.on_round_start(from_java(state)))

    def onRoundEnd(self, state):
        return to_java(self.shadow.on_round_end(from_java(state)))

    def onLeave(self, state):
        return to_java(self.shadow.on_leave(from_java(state)))

    def onPartnerProposal(self, state, agent_id, offer):
        return to_java(
            self.shadow.on_partner_proposal(state=from_java(state), agent_id=agent_id, offer=from_java(offer)))

    def onPartnerRefusedToPropose(self, state, agent_id):
        return to_java(self.shadow.on_partner_refused_to_propose(state=from_java(state), agent_id=agent_id))

    def onPartnerResponse(self, state, agent_id, offer, response: int, counter_offer):
        return to_java(
            self.shadow.on_partner_response(state=from_java(state), agent_id=agent_id, outcome=from_java(offer)
                                            , response=SAOResponse(response=ResponseType(response)
                                                                       , outcome=from_java(counter_offer))))

    def onNotification(self, notification, notifier):
        return to_java(self.shadow.on_notification(from_java(notification), notifier))

    def setUtilityFunction(self, ufun):
        self.shadow.utility_function = ufun if ufun is None else JavaUtilityFunction(ufun, None)

    def getUtilityFunction(self):
        return to_java(self.shadow.utility_function)

    def getID(self):
        return self.shadow.id

    def setID(self):
        return self.shadow.id


class SAOController(Controller):

    def propose_(self, negotiator_id: str, state: MechanismState) -> Optional['Outcome']:
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'propose', state=state)

    def respond_(self, negotiator_id: str, state: MechanismState, offer: 'Outcome') -> 'ResponseType':
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'respond', state=state, offer=offer)


SAOProtocol = SAOMechanism
"""An alias for `SAOMechanism object"""



from typing import List, Optional, Tuple, Dict, Any, Union, Callable, Set, Iterable, Collection

import pytest

from negmas import Issue, AspirationNegotiator, LinearUtilityAggregationFunction, Mechanism, Negotiator, \
    SAOMechanism, Mechanism, Negotiator, RenegotiationRequest, AgentMechanismInterface, MechanismState
from negmas.helpers import ConfigReader
from negmas.situated import World, Agent, Action, Breach, Contract


results = [] # will keep results not to use printing


class DummyWorld(World):
    def _complete_contract_execution(self, contract: Contract, breaches: List[Breach], resolved: bool) -> None:
        pass

    def _contract_finalization_time(self, contract: Contract) -> int:
        return self.current_step + 1

    def _contract_execution_time(self, contract: Contract) -> int:
        return self.current_step

    def _contract_size(self, contract: Contract) -> float:
        return 0.0

    def __init__(self, n_steps=10000, negotiation_speed=None):
        super().__init__(n_steps=n_steps, negotiation_speed=negotiation_speed)
        self.the_agents = []

    def join(self, x: 'Agent', simulation_priority: int = 0):
        super().join(x=x, simulation_priority=simulation_priority)
        self.the_agents.append(x)

    def _delete_executed_contracts(self) -> None:
        pass

    def _get_executable_contracts(self) -> Collection[Contract]:
        return []

    def _post_step_stats(self):
        pass

    def _pre_step_stats(self):
        pass

    def _contract_execution_order(self, contracts: Collection[Contract]) -> Collection[Contract]:
        return contracts

    def _contract_record(self, contract: Contract) -> Dict[str, Any]:
        return contract.__dict__

    def _execute_contract(self, contract: Contract) -> Set[Breach]:
        return set()

    def _process_breach(self, contract: Contract, breaches: List[Breach]
                        , force_immediate_signing=True) -> Optional[Contract]:
        return None

    def _breach_record(self, breach: Breach):
        return breach.__dict__

    def execute(self, action: Action, agent: 'Agent', callback: Callable = None) -> bool:
        return True

    def get_private_state(self, agent: 'Agent') -> Any:
        s = {'partners': [_ for _ in self.the_agents if _ is not agent]}
        return s

    def _simulation_step(self):
        pass


class DummyAgent(Agent):
    def init(self):
        pass

    def _respond_to_negotiation_request(self, initiator: str, partners: List[str], issues: List[Issue],
                                        annotation: Dict[str, Any], mechanism: AgentMechanismInterface,
                                        role: Optional[str], req_id: Optional[str]) -> Optional[Negotiator]:
        negotiator = AspirationNegotiator(
            ufun=LinearUtilityAggregationFunction(issue_utilities=[lambda x: 1.0 - x / 10.0]))
        return negotiator

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        pass

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        pass

    def on_negotiation_failure(self, partners: List[str], annotation: Dict[str, Any],
                               mechanism: AgentMechanismInterface, state: MechanismState) -> None:
        pass

    def on_negotiation_success(self, contract: Contract, mechanism: AgentMechanismInterface) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        pass

    def set_renegotiation_agenda(self, contract: Contract, breaches: List[Breach]) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Breach]
                                         , agenda: RenegotiationRequest) -> Optional[Negotiator]:
        return None

    def __init__(self, name=None):
        super().__init__(name=name)
        self.id = name
        self.current_step = 0

    def step(self):
        global results
        self.current_step = self.awi.current_step
        if (self.current_step == 2 and self.name.endswith('1')) or (self.current_step == 4 and self.name.endswith('2')):
            issues = [Issue(10)]
            partners = self.awi.state['partners']
            self._request_negotiation(partners=[_.name for _ in partners] + [self.name], issues=issues)
            results.append(f'{self.name} started negotiation with {partners[0].name}')
        results.append(f'{self.name}: step {self.current_step}')


def test_world_runs_with_some_negs(capsys):
    global results
    results = []
    world = DummyWorld(n_steps=10)
    world.join(DummyAgent("A1"))
    world.join(DummyAgent("A2"))
    world.run()
    assert "A1: step 1" in results, 'first step logged'
    assert "A2: step 1" in results, 'first step logged'
    assert "A1 started negotiation with A2" in results, 'negotiation started'
    assert f"A1: step {world.n_steps-1}" in results, 'last step logged'
    assert f"A2: step {world.n_steps-1}" in results, 'last step logged'
    assert len(world._saved_contracts) == 2
    assert sum(world.stats['n_negotiations']) == 2


def test_config_reader_with_a_world():

    world = DummyWorld()
    assert world.bulletin_board is not None
    assert world.n_steps == 10000

    world = DummyWorld.from_config(scope=globals(), config={'n_steps': 10})
    assert world.bulletin_board is not None
    assert world.n_steps == 10


def test_config_reader_with_a_world_with_enum():
    from negmas.situated import BulletinBoard, World

    world = DummyWorld()
    assert world.bulletin_board is not None
    assert world.n_steps == 10000

    world = DummyWorld.from_config(scope=globals()
                              , config={'n_steps': 10, 'negotiation_speed': 2})
    assert world.bulletin_board is not None
    assert world.n_steps == 10
    assert world.negotiation_speed == 2


if __name__ == '__main__':
    pytest.main(args=[__file__])



"""
Tournament generation and management.

"""
import concurrent.futures as futures
import copy
import itertools
import math
import pathlib
import time
import traceback
import warnings
from dataclasses import dataclass, field
from os import PathLike
from pathlib import Path
from typing import Optional, List, Callable, Union, Type, Sequence, Dict, Any, Tuple

import numpy as np
import pandas as pd
import yaml
from typing_extensions import Protocol

from negmas.helpers import get_class, unique_name, import_by_name, get_full_type_name, humanize_time, dump
from .situated import Agent, World, save_stats

__all__ = [
    'tournament',
    'WorldGenerator',
    'WorldRunResults',
    'TournamentResults',
    'run_world',
    'process_world_run',
]

PROTOCOL_CLASS_NAME_FIELD = '__mechanism_class_name'

try:
    # disable a warning in yaml 1b1 version
    yaml.warnings({'YAMLLoadWarning': False})
except:
    pass


class WorldGenerator(Protocol):
    """A callback-protocol specifying the signature of a world generator function that can be passed to `tournament`

    Args:
            kwargs: key-value pairs of arguments.

    See Also:
        `tournament`

    """

    def __call__(self, **kwargs) -> World: ...


class ConfigGenerator(Protocol):
    """A callback-protocol specifying the signature of a config generator function that can be passed to `tournament`

    Args:

            n_competitors: Number of competitor types
            n_agents_per_competitor: Number of agents to instantiate for each competitor
            agent_names_reveal_type: whether agent names contain their types (used for debugging purposes).
            kwargs: key-value pairs of arguments.

    See Also:

        `tournament` `ConfigAssigner`

    """

    def __call__(self, n_competitors: int, n_agents_per_competitor: int, agent_names_reveal_type: bool = False
                 , compact: bool = False, **kwargs) -> Dict[str, Any]: ...


class ConfigAssigner(Protocol):
    """A callback-protocol specifying the signature of a function that can be used to assign competitors to a config
     generated using a `ConfigGenerator`

     Args:

         config: The dict returned from the `ConfigGenerator`
         max_n_worlds: Maximum allowed number of worlds to generate from this config
         n_agents_per_competitor: Number of agents to instantiate for each competitor
         competitors: A list of `Agent` types that can be used to create the competitors
         params: A list of parameters to pass to the agent types

     See Also:

         `ConfigGenerator` `tournament`

    """

    def __call__(self, config: Dict[str, Any], max_n_worlds: int, n_agents_per_competitor: int = 1
                 , competitors: Sequence[Type[Agent]] = (), params: Sequence[Dict[str, Any]] = ()
                 ) -> List[Dict[str, Any]]: ...


@dataclass
class WorldRunResults:
    """Results of a world run"""
    world_name: str
    """World name"""
    log_file_name: str
    """Log file name"""
    names: List[str] = field(default_factory=list, init=False)
    """Agent names"""
    scores: List[float] = field(default_factory=list, init=False)
    """Agent scores"""
    types: List[str] = field(default_factory=list, init=False)
    """Agent type names"""


@dataclass
class TournamentResults:
    scores: pd.DataFrame
    total_scores: pd.DataFrame
    winners: List[str]
    """Winner type name(s) which may be a list"""
    winners_scores: np.array
    """Winner score (accumulated)"""
    ttest: pd.DataFrame


def run_world(world_params: dict, dry_run: bool = False, save_world_stats: bool = True):
    """Runs a world and returns stats. This function is designed to be used with distributed systems like dask.

    Args:
        world_params: World info dict. See remarks for its parameters
        dry_run: If true, the world will not be run. Only configs will be saved
        save_world_stats: If true, saves individual world stats

    Remarks:

        The `world_params` dict should have the following members:

            - name: world name [Defaults to random]
            - log_file_name: file name to store the world log [Defaults to random]
            - __dir_name: directory to store the world stats [Defaults to random]
            - __world_generator: full name of the world generator function (including its module) [Required]
            - __score_calculator: full name of the score calculator function [Required]
            - __tournament_name: name of the tournament [Defaults to random]
            - others: values of all other keys are passed to the world generator as kwargs
    """
    world_generator = world_params.get('__world_generator', None)
    score_calculator = world_params.get('__score_calculator', None)
    tournament_name = world_params.get('__tournament_name', unique_name(base=""))
    assert world_generator and score_calculator, f'Cannot run without specifying both a world generator and a score ' \
        f'calculator'

    world_generator = import_by_name(world_generator)
    score_calculator = import_by_name(score_calculator)
    default_name = unique_name(base="")
    world_params['name'] = world_params.get('name', default_name)
    world_name = world_params['name']
    default_dir = (Path(f'~') / 'negmas' / 'tournaments' / tournament_name / world_name).absolute()
    world_params['log_file_name'] = world_params.get('log_file_name', str(default_dir / 'log.txt'))
    world_params['__dir_name'] = world_params.get('__dir_name', str(default_dir))
    # delete the parameters not used by _run_world
    for k in ('__world_generator', '__tournament_name', '__score_calculator'):
        if k in world_params.keys():
            world_params.pop(k, None)
    return _run_world(world_params=world_params, world_generator=world_generator, score_calculator=score_calculator
                      , dry_run=dry_run, save_world_stats=save_world_stats)


def _run_world(world_params: dict, world_generator: WorldGenerator
               , score_calculator: Callable[[World, bool], WorldRunResults]
               , world_progress_callback: Callable[[Optional[World]], None] = None, dry_run: bool = False
               , save_world_stats: bool = True
               ) -> Tuple[WorldRunResults, str]:
    """Runs a world and returns stats

    Args:
        world_params: World info dict. See remarks for its parameters
        world_generator: World generator function.
        score_calculator: Score calculator function.
        world_progress_callback: world progress callback
        dry_run: If true, the world is not run. Its config is saved instead.
        save_world_stats: If true, saves individual world stats

    Returns:

        A tuple with the list of `WorldRunResults` for all the worlds generated using this config and the directory in
        which these results are stored

    Remarks:

        - The `world_params` dict should have the following members:

            - name: world name
            - log_file_name: file name to store the world log
            - __dir_name: directory to store the world stats
            - others: values of all other keys are passed to the world generator as kwargs


    """
    world_params = world_params.copy()
    dir_name = world_params['__dir_name']
    world_params.pop('__dir_name', None)
    world = world_generator(**world_params)
    if dry_run:
        world.save_config(dir_name)
        scores = score_calculator(world, True)
        return scores, dir_name
    if world_progress_callback is None:
        world.run()
    else:
        _start_time = time.monotonic()
        for _ in range(world.n_steps):
            if world.time_limit is not None and (time.monotonic() - _start_time) >= world.time_limit:
                break
            if not world.step():
                break
            world_progress_callback(world)
    if save_world_stats:
        save_stats(world=world, log_dir=dir_name)
    scores = score_calculator(world, False)
    return scores, dir_name


def process_world_run(results: WorldRunResults, tournament_name: str, dir_name: str
                      , save_world_stats: bool = True) -> pd.DataFrame:
    """
    Generates a dataframe with the results of this world run

    Args:
        results: Results of the world run
        tournament_name: tournament name
        dir_name: directory name to store the stats.
        save_world_stats: It True, it will be assumed that world stats are saved

    Returns:

        A pandas DataFrame with agent_name, agent_type, score, log_file, world, and stats_folder columns

    """
    log_file, world_name_ = results.log_file_name, results.world_name
    if save_world_stats and log_file is not None and pathlib.Path(log_file).exists():
        with open(log_file, 'a') as f:
            f.write(f'\nPART of TOURNAMENT {tournament_name}. This world run completed successfully\n')
    scores = []
    for name_, type_, score in zip(results.names, results.types, results.scores):
        scores.append({'agent_name': name_, 'agent_type': type_, 'score': score, 'log_file': log_file
                       , 'world': world_name_, 'stats_folder': dir_name})
    return pd.DataFrame(data=scores)


def _run_dask(scheduler_ip, scheduler_port, verbose, world_infos, world_generator, tournament_progress_callback
              , n_worlds, name, score_calculator, dry_run, save_world_stats) -> List[pd.DataFrame]:
    """Runs the tournament on dask"""

    import distributed
    scores = []
    if scheduler_ip is None and scheduler_port is None:
        address = None
    else:
        if scheduler_ip is None:
            scheduler_ip = '127.0.0.1'
        if scheduler_port is None:
            scheduler_port = '8786'
        address = f'{scheduler_ip}:{scheduler_port}'
    if verbose:
        print(f'Will use DASK on {address}')
    client = distributed.Client(address=address, set_as_default=True)
    future_results = []
    for world_info in world_infos:
        future_results.append(client.submit(_run_world, world_info, world_generator, score_calculator, None, dry_run
                                            , save_world_stats))
    print(f'Submitted all processes to DASK ({len(world_infos)})')
    _strt = time.perf_counter()
    for i, (future, result) in enumerate(
        distributed.as_completed(future_results, with_results=True, raise_errors=False)):
        try:
            score_, dir_name = result
            if tournament_progress_callback is not None:
                tournament_progress_callback(score_, i, n_worlds)
            scores.append(process_world_run(score_, tournament_name=name, dir_name=str(dir_name))
                          , save_world_stats=save_world_stats)
            if verbose:
                _duration = time.perf_counter() - _strt
                print(f'{i + 1:003} of {n_worlds:003} [{100 * (i + 1) / n_worlds:0.3}%] completed in '
                      f'{humanize_time(_duration)} [ETA {humanize_time(_duration * n_worlds / (i + 1))}]')
        except Exception as e:
            if tournament_progress_callback is not None:
                tournament_progress_callback(None, i, n_worlds)
            print(traceback.format_exc())
            print(e)
    client.shutdown()
    return scores


def tournament(competitors: Sequence[Union[str, Type[Agent]]]
               , config_generator: ConfigGenerator
               , config_assigner: ConfigAssigner
               , world_generator: WorldGenerator
               , score_calculator: Callable[[World, bool], WorldRunResults]
               , competitor_params: Optional[Sequence[Dict[str, Any]]] = None
               , agent_names_reveal_type=False
               , n_agents_per_competitor=1
               , n_configs: int = 10
               , max_worlds_per_config: int = 100
               , n_runs_per_world: int = 5
               , max_n_configs: int = None
               , n_runs_per_config: int = None
               , tournament_path: str = './logs/tournaments'
               , total_timeout: Optional[int] = None
               , parallelism='parallel'
               , scheduler_ip: Optional[str] = None
               , scheduler_port: Optional[str] = None
               , tournament_progress_callback: Callable[[Optional[WorldRunResults], int, int], None] = None
               , world_progress_callback: Callable[[Optional[World]], None] = None
               , name: str = None
               , verbose: bool = False
               , configs_only: bool = False
               , compact: bool = False
               , **kwargs
               ) -> Union[TournamentResults, PathLike]:
    """
    Runs a tournament

    Args:

        name: Tournament name
        config_generator: Used to generate unique configs that will be used to evaluate competitors
        config_assigner: Used to generate assignments of competitors to the configs created by the `config_generator`
        world_generator: A functions to generate worlds for the tournament that follows the assignments made by the
                         `config_assigner`
        score_calculator: A function for calculating the score of all agents in a world *After it finishes running*.
                          The second parameter is a boolean specifying whether this is a dry_run. For dry runs, scores
                          are not expected but names and types should exist in the returned `WorldRunResults`.
        competitors: A list of class names for the competitors
        competitor_params: A list of competitor parameters (used to initialize the competitors).
        agent_names_reveal_type: If true then the type of an agent should be readable in its name (most likely at its
                                 beginning).
        n_configs: The number of different world configs (up to competitor assignment) to be generated.
        max_worlds_per_config: The maximum number of worlds to run per config. If None, then all possible assignments
                             of competitors within each config will be tried (all permutations).
        n_runs_per_world: Number of runs per world. All of these world runs will have identical competitor assignment
                          and identical world configuration.
        n_agents_per_competitor: The number of agents of each competing type to be instantiated in the world.
        max_n_configs: [Depricated] The number of configs to use (it is replaced by separately setting `n_config`
                       and `max_worlds_per_config` )
        n_runs_per_config: [Depricated] The number of runs (simulation) for every config. It is replaced by
                           `n_runs_per_world`
        total_timeout: Total timeout for the complete process
        tournament_path: Path at which to store all results. A scores.csv file will keep the scores and logs folder will
                         keep detailed logs
        parallelism: Type of parallelism. Can be 'serial' for serial, 'parallel' for parallel and 'distributed' for
                     distributed!
        scheduler_port: Port of the dask scheduler if parallelism is dask, dist, or distributed
        scheduler_ip:   IP Address of the dask scheduler if parallelism is dask, dist, or distributed
        world_progress_callback: A function to be called after every step of every world run (only allowed for serial
                                 and parallel evaluation and should be used with cautious).
        tournament_progress_callback: A function to be called with `WorldRunResults` after each world finished
                                      processing
        verbose: Verbosity
        configs_only: If true, a config file for each
        compact: If true, compact logs will be created and effort will be made to reduce the memory footprint
        kwargs: Arguments to pass to the `config_generator` function

    Returns:
        `TournamentResults` The results of the tournament or a `PathLike` giving the location where configs were saved

    """
    if max_n_configs is not None or n_runs_per_config is not None:
        n_runs_per_world = n_runs_per_config if n_runs_per_config is not None else n_runs_per_world
        n_configs = max(1, int(math.log2(max_n_configs)))
        max_worlds_per_config = int(0.5 + max_n_configs / n_configs)

        warnings.warn(f'max_n_configs and n_runs_per_config are deprecated and will be removed in future versions. '
                      f'Use n_configs, max_worlds_per_config n_runs_per_world instead.'
                      f'\nWill use the following settings: n_configs ({n_configs})'
                      f', max_worlds_per_config ({max_worlds_per_config})'
                      f', and n_runs_per_world ({n_runs_per_world}).')

    if n_runs_per_world is None or n_configs is None or max_worlds_per_config is None:
        raise ValueError(f'Values for n_configs ({n_configs}), max_worlds_per_config ({max_worlds_per_config})'
                         f', and n_runs_per_world ({n_runs_per_world}) must all be given or possible to calculate '
                         f'from max_n_configs ({max_n_configs}) and n_runs_per_config ({n_runs_per_config})')
    dask_options = ('dist', 'distributed', 'dask', 'd')
    multiprocessing_options = ('local', 'parallel', 'par', 'p')
    serial_options = ('none', 'serial', 's')
    if parallelism is None:
        parallelism = 'serial'
    assert total_timeout is None or parallelism not in dask_options, f'Cannot use {parallelism} with a total-timeout'
    assert world_progress_callback is None or parallelism not in dask_options, f'Cannot use {parallelism} with a ' \
        f'world callback'
    if name is None:
        name = unique_name('', add_time=True, rand_digits=0)
    competitors = list(competitors)
    if tournament_path.startswith('~'):
        tournament_path = Path.home() / ('/'.join(tournament_path.split('/')[1:]))
    tournament_path = (pathlib.Path(tournament_path) / name).absolute()
    tournament_path.mkdir(parents=True, exist_ok=True)
    if verbose:
        print(f'Results of Tournament {name} will be saved to {str(tournament_path)}')
    if competitor_params is None:
        competitor_params = [dict() for _ in range(len(competitors))]
    params = {
        'competitors': [get_class(_).__name__ if not isinstance(_, str) else _ for _ in competitors],
        'params': competitor_params,
        'n_agents_per_competitor': n_agents_per_competitor,
        'tournament_path': str(tournament_path),
        'total_timeout': total_timeout,
        'parallelism': parallelism,
        'scheduler_ip': scheduler_ip,
        'scheduler_port': scheduler_port,
        'name': name,
        'n_configs': n_configs,
        'n_world_per_config': max_worlds_per_config,
        'n_runs_per_world': n_runs_per_world,
        'n_worlds': None
    }
    params.update(kwargs)
    dump(params, tournament_path / 'params')

    configs = [config_generator(n_competitors=len(competitors), n_agents_per_competitor=n_agents_per_competitor
                                , agent_names_reveal_type=agent_names_reveal_type, compact=compact, **kwargs)
               for _ in range(n_configs)]

    dump(configs, tournament_path / 'base_configs')

    assigned = list(itertools.chain(*[config_assigner(config=c, max_n_worlds=max_worlds_per_config
                                                      , n_agents_per_competitor=n_agents_per_competitor
                                                      , competitors=competitors, params=competitor_params)
                                      for c in configs]))

    params['n_worlds'] = len(assigned) * n_runs_per_world

    dump(params, tournament_path / 'params')
    dump(assigned, tournament_path / 'assigned_configs')

    assigned = list(itertools.chain(*([assigned] * n_runs_per_world)))

    for config in assigned:
        dir_name = tournament_path / config['world_params']['name']
        config.update({'log_file_name': str(dir_name / 'log.txt'), '__dir_name': str(dir_name)})
        config['world_params'].update({'log_file_name': str(dir_name / 'log.txt')})

    saved_configs = [{k: copy.copy(v) if k != 'competitors' else
    [get_full_type_name(c) if not isinstance(c, str) else c for c in v]
                      for k, v in _.items()} for _ in assigned]
    score_calculator_name = get_full_type_name(score_calculator) if not isinstance(score_calculator,
                                                                                   str) else score_calculator
    world_generator_name = get_full_type_name(world_generator) if not isinstance(world_generator,
                                                                                 str) else world_generator
    for d in saved_configs:
        d['__score_calculator'] = score_calculator_name
        d['__world_generator'] = world_generator_name
        d['__tournament_name'] = name
    config_path = tournament_path / 'configs'
    config_path.mkdir(exist_ok=True, parents=True)
    for i, conf in enumerate(saved_configs):
        f_name = config_path / f'{i:06}'
        dump(conf, f_name)

    scores_list = []
    scores_file = str(tournament_path / 'scores.csv')
    n_world_configs = len(assigned)

    if verbose:
        print(f'Will run {n_world_configs}  total world simulations ({parallelism})')
    if parallelism in serial_options:
        strt = time.perf_counter()
        for i, world_params in enumerate(assigned):
            if total_timeout is not None and time.perf_counter() - strt > total_timeout:
                break
            try:
                score_, _ = _run_world(world_params=world_params, world_generator=world_generator
                                       , world_progress_callback=world_progress_callback
                                       , score_calculator=score_calculator, dry_run=configs_only
                                       , save_world_stats=not compact)
                if tournament_progress_callback is not None:
                    tournament_progress_callback(score_, i, n_world_configs)
                scores_list.append(process_world_run(score_, tournament_name=name
                                                     , dir_name=str(world_params['__dir_name'])
                                                     , save_world_stats=not compact))
                if verbose:
                    _duration = time.perf_counter() - strt
                    print(f'{i + 1:003} of {n_world_configs:003} [{(i + 1) / n_world_configs:.02%}] '
                          f'{"saved" if configs_only else "completed"} '
                          f'in {humanize_time(_duration)}'
                          f' [ETA {humanize_time(_duration * n_world_configs / (i + 1))}]')
            except Exception as e:
                if tournament_progress_callback is not None:
                    tournament_progress_callback(None, i, n_world_configs)
                print(traceback.format_exc())
                print(e)
    elif parallelism in multiprocessing_options:
        executor = futures.ProcessPoolExecutor(max_workers=None)
        future_results = []
        for world_params in assigned:
            future_results.append(executor.submit(_run_world, world_params, world_generator, score_calculator
                                                  , world_progress_callback, configs_only, not compact))
        if verbose:
            print(f'Submitted all processes ({len(assigned)})')
        _strt = time.perf_counter()
        for i, future in enumerate(futures.as_completed(future_results, timeout=total_timeout)):
            try:
                score_, dir_name = future.result()
                if tournament_progress_callback is not None:
                    tournament_progress_callback(score_, i, n_world_configs)
                scores_list.append(process_world_run(score_, tournament_name=name, dir_name=str(dir_name)
                                                     , save_world_stats=not compact))
                if verbose:
                    _duration = time.perf_counter() - _strt
                    print(f'{i + 1:003} of {n_world_configs:003} [{100 * (i + 1) / n_world_configs:0.3}%] '
                          f'{"saved" if configs_only else "completed"} in '
                          f'{humanize_time(_duration)}'
                          f' [ETA {humanize_time(_duration * n_world_configs / (i + 1))}]')
            except futures.TimeoutError:
                if tournament_progress_callback is not None:
                    tournament_progress_callback(None, i, n_world_configs)
                print('Tournament timed-out')
                break
            except Exception as e:
                if tournament_progress_callback is not None:
                    tournament_progress_callback(None, i, n_world_configs)
                print(traceback.format_exc())
                print(e)
    elif parallelism in dask_options:
        scores_list = _run_dask(scheduler_ip, scheduler_port, verbose, assigned, world_generator
                                , tournament_progress_callback, n_world_configs, name, score_calculator, configs_only
                                , not compact)
    if configs_only:
        return config_path
    if verbose:
        print(f'Finding winners')
    if len(scores_list) < 1:
        return TournamentResults(scores=pd.DataFrame(), total_scores=pd.DataFrame()
                                 , winners=[], winners_scores=np.array([]), ttest=pd.DataFrame())
    scores: pd.DataFrame = pd.concat(scores_list, ignore_index=True)
    scores = pd.DataFrame(data=scores)
    scores.to_csv(scores_file, index_label='index')
    scores = scores.loc[~scores['agent_type'].isnull(), :]
    scores = scores.loc[scores['agent_type'].str.len() > 0, :]
    total_scores = scores.groupby(['agent_type'])['score'].mean().sort_values(ascending=False).reset_index()
    winner_table = total_scores.loc[total_scores['score'] == total_scores['score'].max(), :]
    winners = winner_table['agent_type'].values.tolist()
    winner_scores = winner_table['score'].values
    types = list(scores['agent_type'].unique())

    ttest_results = []
    for i, t1 in enumerate(types):
        for j, t2 in enumerate(types[i + 1:]):
            from scipy.stats import ttest_ind
            t, p = ttest_ind(scores[scores['agent_type'] == t1].score, scores[scores['agent_type'] == t2].score)
            ttest_results.append({'a': t1, 'b': t2, 't': t, 'p': p})

    if verbose:
        print(f'Tournament completed successfully\nWinners: {list(zip(winners, winner_scores))}')

    scores.to_csv(str(tournament_path / 'scores.csv'), index_label='index')
    total_scores.to_csv(str(tournament_path / 'total_scores.csv'), index_label='index')
    winner_table.to_csv(str(tournament_path / 'winners.csv'), index_label='index')
    ttest_results = pd.DataFrame(data=ttest_results)
    ttest_results.to_csv(str(tournament_path / 'ttest.csv'), index_label='index')

    return TournamentResults(scores=scores, total_scores=total_scores, winners=winners, winners_scores=winner_scores
                             , ttest=ttest_results)



from typing import List

import numpy as np
import pytest

from negmas import ToughNegotiator, SAOMechanism, \
    AspirationNegotiator, OnlyBestNegotiator, SAOController
from negmas.utilities import RandomUtilityFunction


def test_tough_asp_negotiator():
    a1 = ToughNegotiator(dynamic_ufun=False)
    a2 = AspirationNegotiator(dynamic_ufun=False, aspiration_type='conceder')
    outcomes = [(_,) for _ in range(10)]
    u1 = np.linspace(0.0, 1.0, len(outcomes))
    u2 = 1.0 - u1
    neg = SAOMechanism(outcomes=outcomes, n_steps=20)
    neg.add(a1, ufun=u1)
    neg.add(a2, ufun=u2)
    neg.run()
    a1offers = [s.current_offer for s in neg.history if s.current_proposer == a1.id]
    a2offers = [s.current_offer for s in neg.history if s.current_proposer == a2.id]
    assert a1._offerable_outcomes is None
    if len(a1offers) > 0:
        assert len(set(a1offers)) == 1 and a1offers[-1] == (9,)
    assert len(set(a2offers)) >= 0


def test_best_only_asp_negotiator():
    a1 = OnlyBestNegotiator(dynamic_ufun=False, min_utility=0.9, top_fraction=0.1)
    a2 = AspirationNegotiator(dynamic_ufun=False, aspiration_type='conceder')
    outcomes = [(_,) for _ in range(20)]
    u1 = np.linspace(0.0, 1.0, len(outcomes))
    u2 = 1.0 - u1
    neg = SAOMechanism(outcomes=outcomes, n_steps=200)
    neg.add(a1, ufun=u1)
    neg.add(a2, ufun=u2)
    neg.run()
    a1offers = [s.current_offer for s in neg.history if s.current_proposer == a1.id]
    a2offers = [s.current_offer for s in neg.history if s.current_proposer == a2.id]
    assert a1._offerable_outcomes is None
    if len(a1offers) > 0:
        assert len(set(a1offers)) <= 2 and min([u1[_[0]] for _ in a1offers if _ is not None]) >= 0.9
    assert len(set(a2offers)) >= 1


def test_controller():
    n_sessions = 5
    c = SAOController(default_negotiator_type='negmas.sao.AspirationNegotiator'
                      , default_negotiator_params={'aspiration_type': 'conceder'})
    sessions = [SAOMechanism(outcomes=10, n_steps=10) for _ in range(n_sessions)]
    for session in sessions:
        session.add(AspirationNegotiator(aspiration_type='conceder')
                    , ufun=RandomUtilityFunction(outcomes=session.outcomes))
        session.add(c.create_negotiator(), ufun=RandomUtilityFunction(outcomes=session.outcomes))
    completed: List[int] = []
    while len(completed) < n_sessions:
        for i, session in enumerate(sessions):
            if i in completed:
                continue
            state = session.step()
            if state.broken or state.timedout or state.agreement is not None:
                completed.append(i)
    # we are just asserting that the controller runs


if __name__ == '__main__':
    pytest.main(args=[__file__])



"""This module defines the interfaces to all negotiation agents (negotiators) in the platform.

"""
import functools
import math
from abc import ABC
from copy import copy
from typing import Optional, Tuple, Union, Type
from typing import TYPE_CHECKING, Dict, Any, Callable

from negmas.common import *
from negmas.events import Notifiable, Notification
from negmas.helpers import get_class
from negmas.utilities import make_discounted_ufun

if TYPE_CHECKING:
    from negmas.outcomes import Outcome
    from negmas.utilities import UtilityValue, UtilityFunction

__all__ = [
    'Negotiator',  # Most abstract kind of agent
    'AspirationMixin',
    'Controller',
]


class Negotiator(NamedObject, Notifiable, ABC):
    r"""Abstract negotiation agent. Base class for all negotiators

     Args:

            name: Negotiator name. If not given it is assigned by the system (unique 16 characters).

        Returns:
            bool: True if participating in the given negotiation (or any negotiation if it was None)

        Remarks:

    """

    def __init__(self, name: str = None, ufun: Optional['UtilityFunction'] = None
                 , parent: 'Controller' = None) -> None:
        super().__init__(name=name)
        self.__parent = parent
        self._capabilities = {'enter': True, 'leave': True}
        self.add_capabilities({'evaluate': True, 'compare': True})
        self._mechanism_id = None
        self._ami = None
        self._initial_state = None
        self.utility_function = ufun
        self._init_utility = ufun
        self._role = None

    def __getattribute__(self, item):
        if item in ('id', 'name') or item.startswith('_'):
            return super().__getattribute__(item)
        parent = super().__getattribute__('__dict__').get('_Negotiator__parent', None)
        if parent is None:
            return super().__getattribute__(item)
        attr = getattr(parent, item, None)
        if attr is None:
            return super().__getattribute__(item)
        if isinstance(attr, Callable):
            return functools.partial(attr, negotiator_id=super().__getattribute__('__dict__')['_NamedObject__uuid'])
        return super().__getattribute__(item)

    def before_death(self, cntxt: Dict[str, Any]) -> bool:
        """Called whenever the parent is about to kill this negotiator. It should return False if the negotiator
        does not with to be killed but the controller can still force-kill it"""

    def _dissociate(self):
        self._mechanism_id = None
        self._ami = None
        self.utility_function = self._init_utility
        self._role = None

    def isin(self, negotiation_id: Optional[str]) -> bool:
        """Is that agent participating in the given negotiation?
        Tests if the agent is participating in the given negotiation.

        Args:

            negotiation_id (Optional[str]): The negotiation ID tested. If
             None, it means ANY negotiation

        Returns:
            bool: True if participating in the given negotiation (or any
                negotiation if it was None)

        """
        return self._mechanism_id == negotiation_id

    @property
    def reserved_value(self):
        """Reserved value is what the agent gets if no agreement is reached in the negotiation."""
        if self.utility_function is None:
            return None
        if self.utility_function.reserved_value is not None:
            return self.utility_function.reserved_value
        return self.utility_function(None)

    @property
    def capabilities(self) -> Dict[str, Any]:
        """Agent capabilities"""
        return self._capabilities

    def add_capabilities(self, capabilities: dict) -> None:
        """Adds named capabilities to the agent.

        Args:
            capabilities: The capabilities to be added as a dict

        Returns:
            None

        Remarks:
            It is the responsibility of the caller to be really capable of added capabilities.

        """
        if hasattr(self, '_capabilities'):
            self._capabilities.update(capabilities)
        else:
            self._capabilities = capabilities

            # CALL BACKS

    def join(self, ami: AgentMechanismInterface, state: MechanismState
                 , *, ufun: Optional['UtilityFunction'] = None, role: str = 'agent') -> bool:
        """
        Called by the mechanism when the agent is about to enter a negotiation. It can prevent the agent from entering

        Args:
            ami  (AgentMechanismInterface): The negotiation.
            state (MechanismState): The current state of the negotiation
            ufun (UtilityFunction): The ufun function to use before any discounting.
            role (str): role of the agent.

        Returns:
            bool indicating whether or not the agent accepts to enter.
            If False is returned it will not enter the negotiation

        """
        if self._mechanism_id is not None:
            return False
        self._role = role
        self._mechanism_id = ami.id
        self._ami = ami
        self._initial_state = state
        if ufun is not None:
            self.utility_function = ufun
        if self.utility_function:
            self.utility_function.ami = ami
        return True

    def on_negotiation_start(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation start

        Args:
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action.

        """

    def on_round_start(self, state: MechanismState) -> None:
        """A call back called at each negotiation round start

        Args:
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action.

        """

    def on_mechanism_error(self, state: MechanismState) -> None:
        """
        A call back called whenever an error happens in the mechanism. The error and its explanation are accessible in
        `state`

        Args:
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action

        """

    def on_round_end(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation round end

        Args:
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action

        """

    def on_leave(self, state: MechanismState) -> None:
        """A call back called after leaving a negotiation.

        Args:
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - **MUST** call the baseclass `on_leave` using `super`() if you are going to override this.
            - The default behavior is to do nothing.
            - Override this to hook some action

        """
        self._dissociate()

    def on_negotiation_end(self, state: MechanismState) -> None:
        """
        A call back called at each negotiation end

        Args:
            state: `MechanismState` or one of its descendants giving the state at which the negotiation ended.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action

        """

    def on_notification(self, notification: Notification, notifier: str):
        if notifier != self._mechanism_id:
            raise ValueError(f'Notification is coming from unknown {notifier}')
        if notification.type == 'negotiation_start':
            self.on_negotiation_start(state=notification.data)
        elif notification.type == 'round_start':
            self.on_round_start(state=notification.data)
        elif notification.type == 'round_end':
            self.on_round_end(state=notification.data)
        elif notification.type == 'negotiation_end':
            self.on_negotiation_end(state=notification.data)

    def __str__(self):
        return f'{self.name}'

    def compare(
        self,
        first: 'Outcome',
        second: 'Outcome',
    ) -> Optional['UtilityValue']:
        """
        Compares two offers using the `ufun`

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared

        Returns:
            UtilityValue: An estimate of the differences between the two outcomes. It can be a real number between -1, 1
            or a probability distribution over the same range.
        """
        if self.utility_function is None:
            return None
        return self.utility_function.compare(first, second)

    class Java:
        implements = ['jnegmas.negotiators.Negotiator']


class AspirationMixin:
    """Adds aspiration level calculation. This Mixin MUST be used with a `Negotiator` class."""

    def aspiration_init(self
                        , max_aspiration: float, aspiration_type: Union[str, int, float], above_reserved_value=True):
        """

        Args:
            max_aspiration:
            aspiration_type:
            above_reserved_value:
        """
        self.add_capabilities({'aspiration': True})
        self.max_aspiration = max_aspiration
        self.aspiration_type = aspiration_type
        self.e = 1.0
        if isinstance(aspiration_type, int):
            self.e = float(aspiration_type)
        elif isinstance(aspiration_type, float):
            self.e = aspiration_type
        elif aspiration_type == 'boulware':
            self.e = 4.0
        elif aspiration_type == 'linear':
            self.e = 1.0
        elif aspiration_type == 'conceder':
            self.e = 0.25
        else:
            raise ValueError(f'Unknown aspiration type {aspiration_type}')
        self.above_reserved = above_reserved_value

    def aspiration(self, t: float) -> float:
        """
        The aspiration level

        Args:
            t: relative time (a number between zero and one)

        Returns:
            aspiration level
        """
        if t is None:
            raise ValueError(f'Aspiration negotiators cannot be used in negotiations with no time or #steps limit!!')
        if self.e < 1e-7:
            return 0.0
        pmin = self.reserved_value if self.above_reserved and self.reserved_value is not None else 0.0
        return pmin + (self.max_aspiration - pmin) * (1.0 - math.pow(t, self.e))


class Controller(NamedObject):
    """Controls the behavior of multiple negotiators in multiple negotiations

    The controller class MUST implement all methods of the negotiator class it is controlling with one added
    argument negotiator_id (str) which represents ID of the negotiator on which the method is being invoked.


    """

    def __init__(self, default_negotiator_type: Union[str, Type[Negotiator]] = None
                 , default_negotiator_params: Dict[str, Any] = None
                 , name: str = None):
        super().__init__(name=name)
        self._negotiators: Dict[str, Tuple['Negotiator', Dict[str, Any]]] = {}
        if default_negotiator_params is None:
            default_negotiator_params = {}
        if isinstance(default_negotiator_type, str):
            default_negotiator_type = get_class(default_negotiator_type)
        self.__default_negotiator_type = default_negotiator_type
        self.__default_negotiator_params = default_negotiator_params

    def create_negotiator(self, negotiator_type: Union[str, Type[Negotiator]] = None
                          , name: str = None
                          , cntxt: Dict[str, None] = None, **kwargs) -> Negotiator:
        """
        Creates a negotiator passing it the context

        Args:
            negotiator_type: Type of the negotiator to be created
            name: negotiator name
            cntxt: The context to be associated with this negotiator. It will not be passed to the negotiator
            constructor.
            **kwargs: any key-value pairs to be passed to the negotiator constructor

        Returns:

            Negotiator: The negotiator to be controlled

        """
        if negotiator_type is None:
            negotiator_type = self.__default_negotiator_type
        elif isinstance(negotiator_type, str):
            negotiator_type = get_class(negotiator_type)
        if negotiator_type is None:
            raise ValueError('No negotiator type is passed and no default negotiator type is defined for this '
                             'controller')
        new_negotiator = negotiator_type(name=name, parent=self, **self.__default_negotiator_params, **kwargs)
        if new_negotiator is not None:
            self._negotiators[new_negotiator.id] = (new_negotiator, cntxt)
        return new_negotiator

    def call(self, negotiator: Negotiator, method: str, *args, **kwargs):
        """
        Calls the given method on the given negotiator safely without causing recursion. The controller MUST use this
        function to access any callable on the negotiator

        Args:
            negotiator:
            method:
            *args:
            **kwargs:

        Returns:

        """
        negotiator._Negotiator__parent = None
        result = getattr(negotiator, method)(*args, **kwargs)
        negotiator._Negotiator__parent = self
        return result

    def kill_negotiator(self, negotiator_id: str, force: bool = False) -> None:
        """
        Kills the negotiator sending it an `before_death` message.

        Args:
            negotiator_id: The ID of the negotiator to kill.
            force: Whether to kill the negotiator in case it refused to die.

        Remarks:

            - Killing a negotiator amounts to nothing more than removing it form the list of negotiators maintained by
              the controller.

        """
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            return
        response = negotiator.on_kill(cntxt=cntxt)
        if response or force:
            self._negotiators.pop(negotiator_id, None)

    def join(self, negotiator_id: str, ami: AgentMechanismInterface, state: MechanismState
                 , *, ufun: Optional['UtilityFunction'] = None, role: str = 'agent') -> bool:
        """
        Called by the mechanism when the agent is about to enter a negotiation. It can prevent the agent from entering

        Args:
            negotiator_id: The negotiator ID
            ami  (AgentMechanismInterface): The negotiation.
            state (MechanismState): The current state of the negotiation
            ufun (UtilityFunction): The ufun function to use before any discounting.
            role (str): role of the agent.

        Returns:
            bool indicating whether or not the agent accepts to enter.If False is returned it will not enter the
            negotiation.

        """
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'join', ami=ami, state=state, ufun=ufun, role=role)

    def on_negotiation_start(self, negotiator_id: str, state: MechanismState) -> None:
        """
        A call back called at each negotiation start

        Args:
            negotiator_id: The negotiator ID
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action.

        """
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'on_negotiation_start', state=state)

    def on_round_start(self, negotiator_id: str, state: MechanismState) -> None:
        """A call back called at each negotiation round start

        Args:
            negotiator_id: The negotiator ID
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action.

        """
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'on_round_start', state=state)

    def on_mechanism_error(self, negotiator_id: str, state: MechanismState) -> None:
        """
        A call back called whenever an error happens in the mechanism. The error and its explanation are accessible in
        `state`

        Args:
            negotiator_id: The negotiator ID
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action

        """
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'on_mechanism_error', state=state)

    def on_round_end(self, negotiator_id: str, state: MechanismState) -> None:
        """
        A call back called at each negotiation round end

        Args:
            negotiator_id: The negotiator ID
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action

        """
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'on_round_end', state=state)

    def on_leave(self, negotiator_id: str, state: MechanismState) -> None:
        """A call back called after leaving a negotiation.

        Args:
            negotiator_id: The negotiator ID
            state: `MechanismState` giving current state of the negotiation.

        Remarks:
            - **MUST** call the baseclass `on_leave` using `super`() if you are going to override this.
            - The default behavior is to do nothing.
            - Override this to hook some action

        """
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'on_leave', state=state)

    def on_negotiation_end(self, negotiator_id: str, state: MechanismState) -> None:
        """
        A call back called at each negotiation end

        Args:
            negotiator_id: The negotiator ID
            state: `MechanismState` or one of its descendants giving the state at which the negotiation ended.

        Remarks:
            - The default behavior is to do nothing.
            - Override this to hook some action

        """
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'on_negotiation_end', state=state)

    def on_notification(self, negotiator_id: str, notification: Notification, notifier: str):
        negotiator, cntxt = self._negotiators.get(negotiator_id, (None, None))
        if negotiator is None:
            raise ValueError(f'Unknown negotiator {negotiator_id}')
        return self.call(negotiator, 'on_notification', notification=notification, notifier=notifier)

    def __str__(self):
        return f'{self.name}'



from pytest import fixture
from negmas.outcomes import Issue


@fixture()
def bissue():
    return Issue(['be', 'not b'], name='binary')


@fixture()
def hamlet():
    return Issue(['val {}'.format(_) for _ in range(5)], name='THE problem')


@fixture()
def uissue():
    return Issue(['val {}'.format(_) for _ in range(5)])


@fixture()
def cissue():
    return Issue((0.0, 1.0), name='c')


@fixture()
def dissue():
    return Issue(10, name='d')


@fixture()
def sissue():
    return Issue(['val {}'.format(_) for _ in range(5)], name='s')


@fixture()
def issues(cissue, dissue, sissue):
    return [cissue, dissue, sissue]


@fixture()
def int_issues():
    return Issue.generate([5], [10])


@fixture()
def valid_outcome_dict(issues):
    outcome = {}
    for issue in issues:
        outcome[issue.name] = issue.rand()
    return outcome


@fixture()
def invalid_outcome_dict(issues):
    outcome = {}
    for issue in issues:
        outcome[issue.name] = issue.rand_invalid()
    return outcome


@fixture()
def valid_outcome_list(int_issues):
    outcome = []
    for issue in int_issues:
        outcome.append(issue.rand())
    return outcome


@fixture()
def invalid_outcome_list(int_issues):
    outcome = []
    for issue in int_issues:
        outcome.append(issue.rand_invalid())
    return outcome



"""Defines import/export functionality
"""
import functools
import operator
import os
import shutil
from typing import List, Optional, Tuple, Union, Dict, Callable, Iterable

import numpy as np
import pkg_resources

from negmas.generics import ivalues

__all__ = [
    'load_genius_domain',
    'load_genius_domain_from_folder',
    'convert_genius_domain_from_folder',
    'convert_genius_domain',
    'find_domain_and_utility_files',
    'get_domain_issues',
]

import xml.etree.ElementTree as ET
from os import listdir
from negmas import Issue, enumerate_outcomes, make_discounted_ufun
from negmas import Negotiator
from negmas import UtilityFunction
from negmas import SAOMechanism, AspirationNegotiator


def get_domain_issues(domain_file_name: str
                       , force_single_issue=False
                       , max_n_outcomes: int = 1e6
                       , n_discretization: Optional[int] = None
                       , keep_issue_names=True
                       , keep_value_names=True
                       , safe_parsing=False
                       ) -> Union[Dict[str, Issue], List[Issue]]:
    """
    Returns the issues of a given XML domain (Genius Format)

    Args:
        domain_file_name:
        force_single_issue:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        safe_parsing:

    Returns:
        List or Dict of issues

    """
    issues, issues_details, mechanism = None, None, None
    if domain_file_name is not None:
        domain_file_name = str(domain_file_name)
        issues_details, _ = Issue.from_genius(domain_file_name, force_single_issue=False
                                              , keep_issue_names=True, keep_value_names=True, safe_parsing=safe_parsing
                                              , n_discretization=n_discretization)
        if force_single_issue:
            issues, _ = Issue.from_genius(domain_file_name, force_single_issue=force_single_issue
                                          , keep_issue_names=keep_issue_names
                                          , keep_value_names=keep_value_names
                                          , max_n_outcomes=max_n_outcomes
                                          , n_discretization=n_discretization)
            if issues is None:
                return []
        else:
            issues, _ = Issue.from_genius(domain_file_name, force_single_issue=force_single_issue
                                          , keep_issue_names=keep_issue_names, keep_value_names=keep_value_names,
                                          safe_parsing=safe_parsing
                                          , n_discretization=n_discretization)
    return issues if not force_single_issue else [issues]


def load_genius_domain(domain_file_name: str
                       , utility_file_names: Optional[List[str]] = None
                       ,
                       agent_factories: Optional[Union[Callable[[], Negotiator], List[Callable[[], Negotiator]]]] = None
                       , force_single_issue=False
                       , cache_and_discretize_outcomes=False
                       , max_n_outcomes: int = 1e6
                       , n_discretization: Optional[int] = None
                       , keep_issue_names=True
                       , keep_value_names=True
                       , normalize_utilities=True
                       , n_steps=None
                       , time_limit=3 * 60  # GENIUS uses 3min time limit by default
                       , max_n_agents=None
                       , dynamic_entry=True
                       , safe_parsing=False
                       , ignore_reserved=False
                       , ignore_discount=False
                       ) \
    -> Tuple[Optional[SAOMechanism], List[dict], Union[Dict[str, Issue], List[Issue]]]:
    """
    Loads a genius domain, creates appropriate negotiators if necessary

    Args:
        domain_file_name:
        utility_file_names:
        agent_factories:
        force_single_issue:
        cache_and_discretize_outcomes:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        normalize_utilities:
        n_steps:
        time_limit:
        max_n_agents:
        dynamic_entry:
        safe_parsing:
        ignore_reserved:
        ignore_discount:

    Returns:
        - mechanism (SAOMechanism): A mechanism for the given issues
        - agent_info (List[Dict]): All Negotiator functions from the given file
        - issues Union[Issue, Dict[str, Issue], List[Issue]]] : The issues

    """
    issues, issues_details, mechanism = None, None, None
    if domain_file_name is not None:
        domain_file_name = str(domain_file_name)
        issues_details, _ = Issue.from_genius(domain_file_name, force_single_issue=False
                                              , keep_issue_names=True, keep_value_names=True, safe_parsing=safe_parsing
                                              , n_discretization=n_discretization)
        if force_single_issue:
            issues, _ = Issue.from_genius(domain_file_name, force_single_issue=force_single_issue
                                          , keep_issue_names=keep_issue_names
                                          , keep_value_names=keep_value_names
                                          , max_n_outcomes=max_n_outcomes
                                          , n_discretization=n_discretization)
            if issues is None:
                return None, [], []
        else:
            issues, _ = Issue.from_genius(domain_file_name, force_single_issue=force_single_issue
                                          , keep_issue_names=keep_issue_names, keep_value_names=keep_value_names,
                                          safe_parsing=safe_parsing
                                          , n_discretization=n_discretization)

    agent_info = []
    if utility_file_names is None:
        utility_file_names = []
    utility_file_names = [str(_) for _ in utility_file_names]
    for ufname in utility_file_names:
        utility, discount_factor = UtilityFunction.from_genius(file_name=ufname
                                                               , force_single_issue=force_single_issue
                                                               , keep_issue_names=keep_issue_names
                                                               , keep_value_names=keep_value_names
                                                               , normalize_utility=normalize_utilities
                                                               , domain_issues=issues_details
                                                               , safe_parsing=safe_parsing
                                                               , max_n_outcomes=max_n_outcomes
                                                               , ignore_discount=ignore_discount
                                                               , ignore_reserved=ignore_reserved
                                                               )
        agent_info.append({
            'ufun': utility, 'reserved_value_func': utility.reserved_value if utility is not None else 0.0
            , 'discount_factor': discount_factor
        })
    outcomes = None
    try:
        if force_single_issue or cache_and_discretize_outcomes or len(issues) == 1:
            n_outcomes: float = functools.reduce(operator.mul, (float(_.cardinality()) if not _.is_continuous() else
                                                                float(
                                                                    n_discretization) if n_discretization is not None else np.inf
                                                                for _ in ivalues(issues)), 1.0)
            if n_outcomes < max_n_outcomes:
                outcomes = enumerate_outcomes(issues, keep_issue_names=keep_issue_names)
    except ValueError:
        pass
    if domain_file_name is not None:
        mechanism_name = domain_file_name.split('/')[-1][:-4].replace('-domain', '').replace('_domain', ''
                                                                                             ).replace('domain', '')
        mechanism = SAOMechanism(issues=issues
                                 , outcomes=outcomes
                                 , n_steps=n_steps, time_limit=time_limit
                                 , max_n_agents=max_n_agents
                                 , dynamic_entry=dynamic_entry, name=mechanism_name
                                 , keep_issue_names=keep_issue_names
                                 )
        if agent_info is not None and len(agent_info) > 0:
            for info in agent_info:
                info['ufun'] = info['ufun'] if info['discount_factor'] is None or info['discount_factor'] == 1.0 else \
                    make_discounted_ufun(ufun=info['ufun'], ami=mechanism.ami
                                         , discount_per_round=info['discount_factor'], power_per_round=1.0)

    if agent_factories is not None and agent_info is not None and len(agent_info) > 0:
        if not isinstance(agent_factories, Iterable):
            agent_factories = [agent_factories] * len(agent_info)
        agents = [factory() for factory in agent_factories[0:len(agent_info)]]
        for a, info in zip(agents, agent_info):
            mechanism.add(a, ufun=info['ufun'])

    return mechanism, agent_info, (issues if not force_single_issue else [issues])


def load_genius_domain_from_folder(folder_name: str
                                   , agent_factories: Optional[
        Union[Callable[[], Negotiator], List[Callable[[], Negotiator]]]] = None
                                   , force_single_issue=False
                                   , cache_and_discretize_outcomes=False
                                   , max_n_outcomes: int = 1e6
                                   , n_discretization: Optional[int] = None
                                   , keep_issue_names=True
                                   , keep_value_names=True
                                   , normalize_utilities=True
                                   , n_steps=None
                                   , time_limit=60# GENIUS uses 3min time limit by default
                                   , max_n_agents=None
                                   , dynamic_entry=True
                                   , safe_parsing=False
                                   , ignore_reserved=False
                                   , ignore_discount=False
                                   ) \
    -> Tuple[Optional[SAOMechanism], List[dict], Union[Dict[str, Issue], List[Issue]]]:
    """
    Loads a genius domain from a folder. See ``load_genius_domain`` for more details.

    Args:
        folder_name:
        agent_factories:
        force_single_issue:
        cache_and_discretize_outcomes:
        max_n_outcomes:
        n_discretization:
        keep_issue_names:
        keep_value_names:
        normalize_utilities:
        n_steps:
        time_limit:
        max_n_agents:
        dynamic_entry:
        safe_parsing:
        ignore_reserved:
        ignore_discount:

    Returns:
        - mechanism (SAOMechanism): A mechanism for the given issues
        - agent_info (List[Dict]): All Negotiator functions from the given file
        - issues Union[Issue, Dict[str, Issue], List[Issue]]] : The issues

    Examples:

        >>> folder_name = pkg_resources.resource_filename('negmas', resource_name='tests/data/10issues')
        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(folder_name
        ...                             , force_single_issue=False, keep_issue_names=False
        ...                             , keep_value_names=False, normalize_utilities=False)
        >>> print(len(issues))
        10
        >>> print(len(negotiators))
        2
        >>> print([type(a['ufun']) for a in negotiators])
        [<class 'negmas.utilities.HyperRectangleUtilityFunction'>, <class 'negmas.utilities.HyperRectangleUtilityFunction'>]
        >>> print(negotiators[0]['ufun'].outcome_ranges[0])
        {1: (7.0, 9.0), 3: (2.0, 7.0), 5: (0.0, 8.0), 8: (0.0, 7.0)}
        >>> print(negotiators[0]['ufun'].mappings[0])
        97.0
        >>> u = negotiators[0]['ufun']
        >>> print(u(tuple([0.0] * len(issues))))
        0.0
        >>> print(u(tuple([0.5] * len(issues))))
        186.0

        Try loading and running a domain with predetermined agents:
        >>> mechanism, agents, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        ...                             , agent_factories=AspirationNegotiator
        ...                             , force_single_issue=True, keep_issue_names=False
        ...                             , keep_value_names=False)
        >>> state = mechanism.run()
        >>> state.agreement is not None
        True


        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop'))

        >>> len(issues), len(negotiators)
        (3, 2)

        >>> [type(a['ufun']) for a in negotiators]
        [<class 'negmas.utilities.LinearUtilityAggregationFunction'>, <class 'negmas.utilities.LinearUtilityAggregationFunction'>]

        >>> mechanism, negotiators, issues = load_genius_domain_from_folder(
        ...                             pkg_resources.resource_filename('negmas', resource_name='tests/data/Laptop')
        ...                             , force_single_issue=True, keep_issue_names=False
        ...                             , keep_value_names=False)
        >>> len(issues), len(negotiators)
        (1, 2)
        >>> [type(a['ufun']) for a in negotiators]
        [<class 'negmas.utilities.MappingUtilityFunction'>, <class 'negmas.utilities.MappingUtilityFunction'>]

    """
    folder_name = str(folder_name)
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith('.xml') or f.endswith('pareto.xml'):
            continue
        full_name = folder_name + '/' + f
        root = ET.parse(full_name).getroot()

        if root.tag == 'negotiation_template':
            domain_file_name = full_name
        elif root.tag == 'utility_space':
            utility_file_names.append(full_name)
    return load_genius_domain(domain_file_name=domain_file_name
                              , utility_file_names=utility_file_names
                              , agent_factories=agent_factories
                              , force_single_issue=force_single_issue
                              , cache_and_discretize_outcomes=cache_and_discretize_outcomes
                              , max_n_outcomes=max_n_outcomes
                              , n_discretization=n_discretization
                              , keep_issue_names=keep_issue_names
                              , keep_value_names=keep_value_names
                              , normalize_utilities=normalize_utilities
                              , n_steps=n_steps
                              , time_limit=time_limit
                              , max_n_agents=max_n_agents
                              , dynamic_entry=dynamic_entry
                              , safe_parsing=safe_parsing
                              , ignore_reserved=ignore_reserved
                              , ignore_discount=ignore_discount)


def find_domain_and_utility_files(folder_name) -> Tuple[str, List[str]]:
    """Finds the domain and utility_function files in a folder
    """
    files = sorted(listdir(folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith('.xml') or f.endswith('pareto.xml'):
            continue
        full_name = folder_name + '/' + f
        root = ET.parse(full_name).getroot()

        if root.tag == 'negotiation_template':
            domain_file_name = full_name
        elif root.tag == 'utility_space':
            utility_file_names.append(full_name)
    return domain_file_name, utility_file_names


def convert_genius_domain(src_domain_file_name: str, dst_domain_file_name: str
                          , src_utility_file_names: Optional[List[str]] = None
                          , dst_utility_file_names: Optional[List[str]] = None
                          , force_single_issue=False
                          , cache_and_discretize_outcomes=False
                          , max_n_outcomes: int = 1e6
                          , n_discretization: Optional[int] = None
                          , keep_issue_names=True
                          , keep_value_names=True
                          , normalize_utilities=True
                          , safe_parsing=False
                          ) -> bool:
    if not force_single_issue and not cache_and_discretize_outcomes and keep_issue_names and keep_value_names \
        and not normalize_utilities:
        # no need to do anything, just copy
        shutil.copy(src=src_domain_file_name, dst=dst_domain_file_name)
        for src, dst in zip(src_utility_file_names, dst_utility_file_names):
            shutil.copy(src=src, dst=dst)
        return True
    issues, issues_details, mechanism = None, None, None
    if src_domain_file_name is not None:
        issues_details, _ = Issue.from_genius(src_domain_file_name, force_single_issue=False
                                              , keep_issue_names=True, keep_value_names=True, safe_parsing=safe_parsing
                                              , n_discretization=n_discretization)
        if force_single_issue:
            issues, _ = Issue.from_genius(src_domain_file_name, force_single_issue=force_single_issue
                                          , keep_issue_names=keep_issue_names
                                          , keep_value_names=keep_value_names
                                          , max_n_outcomes=max_n_outcomes
                                          , n_discretization=n_discretization)
        else:
            issues = issues_details
        if issues is None:
            return False
        Issue.to_genius(issues=issues, file_name=dst_domain_file_name
                        , enumerate_integer=True)

    if src_utility_file_names is None:
        src_utility_file_names = []
    for ufname, dstfname in zip(src_utility_file_names, dst_utility_file_names):
        utility, discount_factor = UtilityFunction.from_genius(file_name=ufname
                                                               , force_single_issue=force_single_issue
                                                               , keep_issue_names=keep_issue_names
                                                               , keep_value_names=keep_value_names
                                                               , normalize_utility=normalize_utilities
                                                               , domain_issues=issues_details
                                                               , safe_parsing=safe_parsing
                                                               , max_n_outcomes=max_n_outcomes)
        if utility is None:
            return False
        UtilityFunction.to_genius(u=utility, discount_factor=discount_factor
                                  , issues=issues, file_name=dstfname)

    return True


def convert_genius_domain_from_folder(src_folder_name: str, dst_folder_name: str, **kwargs) -> bool:
    """Loads a genius domain from a folder. See ``load_genius_domain`` for more details.



    """
    os.makedirs(dst_folder_name, exist_ok=True)
    files = sorted(listdir(src_folder_name))
    domain_file_name = None
    utility_file_names = []
    for f in files:
        if not f.endswith('.xml') or f.endswith('pareto.xml'):
            continue
        full_name = src_folder_name + '/' + f
        root = ET.parse(full_name).getroot()

        if root.tag == 'negotiation_template':
            domain_file_name = full_name
        elif root.tag == 'utility_space':
            utility_file_names.append(full_name)
    success = convert_genius_domain(src_domain_file_name=domain_file_name
                                    , dst_domain_file_name=os.path.join(dst_folder_name,
                                                                        os.path.basename(domain_file_name))
                                    , src_utility_file_names=utility_file_names
                                    , dst_utility_file_names=[os.path.join(dst_folder_name,
                                                                           os.path.basename(_))
                                                              for _ in utility_file_names]
                                    , **kwargs)

    if not success:
        if len(os.listdir(dst_folder_name)) == 0:
            shutil.rmtree(dst_folder_name, ignore_errors=True)

    return success



"""Implements Event management"""
import itertools
import random
from collections import defaultdict
from typing import Any, Dict, Union, Optional, Set

from dataclasses import dataclass

from negmas import NamedObject

__all__ = [
    "Event",
    "EventSource",
    "EventSink",
    "Notification",
    "Notifier",
    "Notifiable"
]


@dataclass
class Event:
    __slots__ = ['type', 'data']
    type: str
    data: Any


class EventSource:
    """An object capable of raising events"""

    def __init__(self):
        super().__init__()
        self.__sinks: Dict[str, list] = defaultdict(list)

    def announce(self, event: Event):
        """Raises an event and informs all event sinks that are registerd for notifications
        on this event type"""

        sinks = self.__sinks.get(event.type, [])
        random.shuffle(sinks)
        for sink in sinks:
            sink.on_event(event=event, sender=self)

    def register_listener(self, event_type: str, listener: "EventSink"):
        self.__sinks[event_type].append(listener)


class EventSink:
    def on_event(self, event: Event, sender: EventSource):
        pass


@dataclass
class Notification:
    __slots__ = ['type', 'data']
    type: str
    data: Any


class Notifier(NamedObject):
    def notify(self, notifiable: 'Notifiable', notification: Notification):
        notifiable.on_notification(notification=notification, notifier=self.id)


class Notifiable:
    def on_notification(self, notification: Notification, notifier: str) -> None:
        pass



import random

from pytest import mark

from negmas import SAOMechanism, AspirationNegotiator, MappingUtilityFunction


def test_on_negotiation_start():
    mechanism = SAOMechanism(outcomes=10)
    assert mechanism.on_negotiation_start()


@mark.parametrize(['n_negotaitors'], [(2,), (3,)])
def test_round_n_agents(n_negotaitors):
    n_outcomes = 5
    mechanism = SAOMechanism(outcomes=n_outcomes, n_steps=3)
    ufuns = MappingUtilityFunction.generate_random(n_negotaitors, outcomes=n_outcomes)
    for i in range(n_negotaitors):
        mechanism.add(AspirationNegotiator(name=f'agent{i}'), ufun=ufuns[i])
    assert mechanism.state.step == 0
    mechanism.step()
    assert mechanism.state.step == 1
    assert 0 <= mechanism._first_proposer < n_negotaitors
    assert mechanism._current_proposer.name == mechanism.negotiators[mechanism._first_proposer].name
    assert mechanism._current_offer is not None


@mark.parametrize(['n_negotaitors'], [(2,), (3,)])
def test_mechanism_can_run(n_negotaitors):
    n_outcomes = 5
    mechanism = SAOMechanism(outcomes=n_outcomes, n_steps=3)
    ufuns = MappingUtilityFunction.generate_random(n_negotaitors, outcomes=n_outcomes)
    for i in range(n_negotaitors):
        mechanism.add(AspirationNegotiator(name=f'agent{i}'), ufun=ufuns[i])
    assert mechanism.state.step == 0
    mechanism.step()
    mechanism.run()
